"""
Tests the parmed/structure module
"""
from io import StringIO
import bz2
import gzip
from copy import copy
import numpy as np
import os
import parmed as pmd
from parmed.exceptions import CharmmWarning, ParameterWarning
import parmed.structure as structure
from parmed.topologyobjects import *
import parmed.unit as u
from parmed.utils import PYPY
import random
import string
import unittest
from unittest import mock
from utils import create_random_structure, get_fn, FileIOTestCase, has_openmm, app, HAS_GROMACS
import warnings
import pickle
import pytest

class TestStructureAPI(unittest.TestCase):
    """ Tests the underlying Structure API """

    def setUp(self):
        s = self.s = structure.Structure()
        s.add_atom(Atom(atomic_number=6), 'ALA', 1, 'A')
        s.add_atom(Atom(atomic_number=1), 'ALA', 1, 'A')
        s.add_atom(Atom(atomic_number=7), 'ALA', 1, 'A')
        s.add_atom(Atom(atomic_number=8), 'ALA', 1, 'A')
        s.add_atom(Atom(atomic_number=16), 'GLY', 2, 'A')
        s.add_atom(Atom(atomic_number=99), 'GLY', 3, 'A')
        s.add_atom(Atom(atomic_number=6), 'GLY', 3, 'B')
        s.add_atom(Atom(atomic_number=6), 'GLY', 3, 'B')
        s.add_atom(Atom(atomic_number=6), 'GLY', 3, 'B')

    @unittest.skipUnless(has_openmm, 'Cannot test without OpenMM')
    def test_gb_assignment(self):
        """ Tests GB parameter assignment """
        # GBneck1
        params = self.s._get_gb_parameters(app.GBn)
        # Radii are assigned elsewhere -- check screen
        self.assertEqual(params[0][1], 0.48435382330)
        self.assertEqual(params[1][1], 1.09085413633)
        self.assertEqual(params[2][1], 0.700147318409)
        self.assertEqual(params[3][1], 1.06557401132)
        self.assertEqual(params[4][1], 0.602256336067)
        self.assertEqual(params[5][1], 0.5)
        # GBneck2
        self.s.bonds.append(Bond(self.s[0], self.s[1]))
        params = self.s._get_gb_parameters(app.GBn2)
        self.assertEqual(params[0][1], 1.058554)
        self.assertEqual(params[0][2], 0.733756)
        self.assertEqual(params[0][3], 0.506378)
        self.assertEqual(params[0][4], 0.205844)
        self.assertEqual(params[1][1], 1.425952)
        self.assertEqual(params[1][2], 0.788440)
        self.assertEqual(params[1][3], 0.798699)
        self.assertEqual(params[1][4], 0.437334)
        self.assertEqual(params[2][1], 0.733599)
        self.assertEqual(params[2][2], 0.503364)
        self.assertEqual(params[2][3], 0.316828)
        self.assertEqual(params[2][4], 0.192915)
        self.assertEqual(params[3][1], 1.061039)
        self.assertEqual(params[3][2], 0.867814)
        self.assertEqual(params[3][3], 0.876635)
        self.assertEqual(params[3][4], 0.387882)
        self.assertEqual(params[4][1], -0.703469)
        self.assertEqual(params[4][2], 0.867814)
        self.assertEqual(params[4][3], 0.876635)
        self.assertEqual(params[4][4], 0.387882)
        self.assertEqual(params[5][1], 0.5)
        self.assertEqual(params[5][2], 1.0)
        self.assertEqual(params[5][3], 0.8)
        self.assertEqual(params[5][4], 4.85)

    def test_combining_rule(self):
        """ Tests the Structure.combining_rule attribute """
        self.assertEqual(self.s.combining_rule, 'lorentz')
        self.s.combining_rule = 'geometric'
        self.assertEqual(self.s.combining_rule, 'geometric')
        def fail():
            self.s.combining_rule = 'badchoice'
        self.assertRaises(ValueError, fail)

    def test_add_atom(self):
        """ Tests the Structure.add_atom method """
        # Check that we have the expected number of residues and atoms
        s = self.s
        self.assertEqual(len(s.atoms), 9)
        self.assertEqual(len(s.residues), 4)
        self.assertEqual(len(s.residues[0]), 4)
        self.assertEqual(len(s.residues[1]), 1)
        self.assertEqual(len(s.residues[2]), 1)
        self.assertEqual(len(s.residues[3]), 3)

        for residue in s.residues:
            self.assertEqual(residue.atoms[-1].idx - residue.atoms[0].idx + 1,
                             len(residue))

    def test_split_duplicate_detection(self):
        """ Tests detection of duplicate molecules in split """
        parm = pmd.load_file(get_fn('ash.parm7')) * 5
        self.assertEqual(len(parm.split()), 1)
        self.assertEqual(len(parm.split()[0][1]), 5)
        parm.residues[0].name = 'ACE1'
        x = parm.split()
        self.assertEqual(len(x), 2)
        self.assertEqual(len(x[0][1]), 1)
        self.assertEqual(len(x[1][1]), 4)
        parm.residues[0].name = 'ACE'
        self.assertEqual(len(parm.split()), 1)
        parm[0].name += '1'
        x = parm.split()
        self.assertEqual(len(x), 2)
        self.assertEqual(len(x[0][1]), 1)
        self.assertEqual(len(x[1][1]), 4)

    def test_split_single_atom(self):
        """ Tests splitting of a structure with single atom residues """
        s = structure.Structure()
        na = pmd.periodic_table.AtomicNum['Na']
        cl = pmd.periodic_table.AtomicNum['Cl']
        li = pmd.periodic_table.AtomicNum['Li']
        f = pmd.periodic_table.AtomicNum['F']
        s.add_atom(Atom(name='NA', type='Na+', atomic_number=na), 'NA', 1, 'A')
        s.add_atom(Atom(name='NA', type='Na+', atomic_number=na), 'NA', 2, 'A')
        s.add_atom(Atom(name='LI', type='Li+', atomic_number=li), 'LI', 3, 'A')
        s.add_atom(Atom(name='LI', type='Li+', atomic_number=li), 'LI', 4, 'A')
        s.add_atom(Atom(name='CL', type='Cl-', atomic_number=cl), 'CL', 5, 'A')
        s.add_atom(Atom(name='CL', type='Cl-', atomic_number=cl), 'CL', 6, 'A')
        s.add_atom(Atom(name='F', type='F-', atomic_number=f), 'F', 7, 'B')
        s.add_atom(Atom(name='F', type='F-', atomic_number=f), 'F', 8, 'B')
        x = s.split()
        self.assertEqual(len(x), 4)
        self.assertEqual(set(len(_[1]) for _ in x), {2})

    def test_bool(self):
        """ Tests bool-ness of Structure """
        self.assertTrue(bool(self.s))
        self.assertFalse(bool(structure.Structure()))

    def test_add_atom_to_residue(self):
        """ Tests the Structure.add_atom_to_residue method """
        s = self.s
        res = s.residues[1]
        s.add_atom_to_residue(Atom(name='TOK'), res)
        s = self.s
        self.assertEqual(len(s.atoms), 10)
        self.assertEqual(len(s.residues), 4)
        self.assertEqual(len(s.residues[0]), 4)
        self.assertEqual(len(s.residues[1]), 2)
        self.assertEqual(len(s.residues[2]), 1)
        self.assertEqual(len(s.residues[3]), 3)

        for residue in s.residues:
            self.assertEqual(residue.atoms[-1].idx - residue.atoms[0].idx + 1,
                             len(residue))
        self.assertEqual(s.atoms[5].name, 'TOK')

        # Now try to add on to the end
        res = s.residues[-1]
        s.add_atom_to_residue(Atom(name='TOK2'), res)
        self.assertEqual(len(s.atoms), 11)
        self.assertEqual(len(s.residues), 4)
        self.assertEqual(len(s.residues[0]), 4)
        self.assertEqual(len(s.residues[1]), 2)
        self.assertEqual(len(s.residues[2]), 1)
        self.assertEqual(len(s.residues[3]), 4)

        for residue in s.residues:
            self.assertEqual(residue.atoms[-1].idx - residue.atoms[0].idx + 1,
                             len(residue))
        self.assertEqual(s.atoms[5].name, 'TOK')
        self.assertEqual(s.atoms[-1].name, 'TOK2')

        # Error handling
        self.assertRaises(ValueError, lambda:
                s.add_atom_to_residue(Atom(name='AXE'), Residue('ALA')))

    def test_box_handling(self):
        """ Tests that Structure.box is always the expected type """
        s = create_random_structure(parametrized=False)
        self.assertIs(s.box, None)
        self.assertIs(s.get_box(), None)
        self.assertRaises(IndexError, lambda: s.get_box(0))
        s.box = [10, 10, 10, 90, 90, 90]
        self.assertIsInstance(s.box, np.ndarray)
        self.assertEqual(s.box[0], 10)
        self.assertEqual(s.box[1], 10)
        self.assertEqual(s.box[2], 10)
        self.assertEqual(s.box[3], 90)
        self.assertEqual(s.box[4], 90)
        self.assertEqual(s.box[5], 90)
        s.box = np.array([10, 10, 10, 90, 90, 90], dtype=np.float64)
        self.assertIsInstance(s.box, np.ndarray)
        self.assertEqual(s.box[0], 10)
        self.assertEqual(s.box[1], 10)
        self.assertEqual(s.box[2], 10)
        self.assertEqual(s.box[3], 90)
        self.assertEqual(s.box[4], 90)
        self.assertEqual(s.box[5], 90)
        s.box = [10*u.angstroms, 10*u.angstroms, 10*u.angstroms, 90*u.degrees, 90*u.degrees, 90*u.degrees]
        self.assertIsInstance(s.box, np.ndarray)
        self.assertEqual(s.box[0], 10)
        self.assertEqual(s.box[1], 10)
        self.assertEqual(s.box[2], 10)
        self.assertEqual(s.box[3], 90)
        self.assertEqual(s.box[4], 90)
        self.assertEqual(s.box[5], 90)
        s.box = [[10*u.angstroms, 1*u.nanometers, 10*u.angstroms, 90*u.degrees, 90*u.degrees, 90*u.degrees],
                 [11*u.angstroms, 1*u.nanometers, 11*u.angstroms, 91*u.degrees, 91*u.degrees, 91*u.degrees]]
        self.assertIsInstance(s.box, np.ndarray)
        self.assertEqual(s.box[0], 10)
        self.assertEqual(s.box[1], 10)
        self.assertEqual(s.box[2], 10)
        self.assertEqual(s.box[3], 90)
        self.assertEqual(s.box[4], 90)
        self.assertEqual(s.box[5], 90)
        box = s.get_box(0)
        self.assertEqual(box[0], 10)
        self.assertEqual(box[1], 10)
        self.assertEqual(box[2], 10)
        self.assertEqual(box[3], 90)
        self.assertEqual(box[4], 90)
        self.assertEqual(box[5], 90)
        box = s.get_box(1)
        self.assertEqual(box[0], 11)
        self.assertEqual(box[1], 10)
        self.assertEqual(box[2], 11)
        self.assertEqual(box[3], 91)
        self.assertEqual(box[4], 91)
        self.assertEqual(box[5], 91)
        box = s.get_box('all')
        self.assertEqual(box[0][0], 10)
        self.assertEqual(box[0][1], 10)
        self.assertEqual(box[0][2], 10)
        self.assertEqual(box[0][3], 90)
        self.assertEqual(box[0][4], 90)
        self.assertEqual(box[0][5], 90)
        self.assertEqual(box[1][0], 11)
        self.assertEqual(box[1][1], 10)
        self.assertEqual(box[1][2], 11)
        self.assertEqual(box[1][3], 91)
        self.assertEqual(box[1][4], 91)
        self.assertEqual(box[1][5], 91)
        self.assertRaises(IndexError, lambda: s.get_box(3))
        # Now test box vectors
        s.box_vectors = [[1, 0, 0], [0, 1, 0], [0, 0, 1]] * u.nanometers
        np.testing.assert_equal(s.box, [10, 10, 10, 90, 90, 90])

    def test_bad_box_handling(self):
        """ Tests error handling when Structure.box is improperly assigned """
        s = create_random_structure(parametrized=True)
        def wrong_number_of_args():
            s.box = [0, 1, 2, 3, 4]
        def wrong_number_of_args2():
            s.box = [0, 1, 2, 3, 4, 5, 6]
        self.assertRaises(ValueError, wrong_number_of_args)
        self.assertRaises(ValueError, wrong_number_of_args2)
        try:
            wrong_number_of_args()
        except ValueError as err:
            self.assertIn('6', str(err))
        # Try wrong units
        for i in range(6):
            box = [10, 10, 10, 90, 90, 90]
            box[i] *= u.liters
            def func():
                s.box = box
            self.assertRaises(TypeError, func)

    def test_coordinates(self):
        """ Tests coordinate handling in Structure """
        s = create_random_structure(parametrized=False)
        self.assertIs(s.coordinates, None)
        self.assertIs(s.positions, None)
        natom = len(s.atoms)
        # Make sure coordinates will be generated from Atom.xx, xy, xz if not
        # otherwise generated
        xyz = np.random.random((natom, 3))
        for a, x in zip(s.atoms, xyz):
            a.xx, a.xy, a.xz = x
        self.assertEqual(s.coordinates.shape, (natom, 3))
        np.testing.assert_equal(s.coordinates, xyz[:,:])
        self.assertIs(s._coordinates, None)
        # Now set multiple frames
        xyz = np.random.random((5, natom, 3)).tolist()
        s.coordinates = xyz
        self.assertIsInstance(s.coordinates, np.ndarray)
        self.assertEqual(s.coordinates.shape, (natom, 3))
        for a, x in zip(s.atoms, xyz[0]):
            self.assertEqual(a.xx, x[0])
            self.assertEqual(a.xy, x[1])
            self.assertEqual(a.xz, x[2])
        np.testing.assert_equal(s.get_coordinates('all'), xyz)
        for i in range(5):
            np.testing.assert_equal(s.get_coordinates(i), xyz[i])
        # Now try setting with units
        xyz = u.Quantity(np.random.random((3, natom, 3)), u.nanometers)
        s.coordinates = xyz
        self.assertIsInstance(s.coordinates, np.ndarray)
        self.assertEqual(s.coordinates.shape, (natom, 3))
        for a, x in zip(s.atoms, xyz[0]._value):
            self.assertEqual(a.xx, x[0]*10)
            self.assertEqual(a.xy, x[1]*10)
            self.assertEqual(a.xz, x[2]*10)
        # Now check setting None
        s.coordinates = None
        for a in s.atoms:
            self.assertFalse(hasattr(a, 'xx'))
            self.assertFalse(hasattr(a, 'xy'))
            self.assertFalse(hasattr(a, 'xz'))
        self.assertIs(s.coordinates, None)
        # Now check setting flattened arrays
        s.coordinates = np.random.random((natom, 3))
        self.assertEqual(s.coordinates.shape, (natom, 3))
        s.coordinates = np.random.random(natom*3)
        self.assertEqual(s.coordinates.shape, (natom, 3))
        s.coordinates = np.random.random(natom*3*10)
        self.assertEqual(s.coordinates.shape, (natom, 3))
        # Now check other iterables
        old_crds = s.coordinates
        s.coordinates = (random.random() for i in range(3*len(s.atoms)))
        self.assertEqual(s.coordinates.shape, (natom, 3))
        diff = (old_crds - s.coordinates).ravel()**2
        self.assertGreater(diff.sum(), 0.01)
        # Check setting the positions attribute
        new_crds = np.random.rand(len(s.atoms), 3) * u.nanometers
        s.positions = new_crds
        np.testing.assert_allclose(new_crds*10, s.coordinates)
        new_crds2 = np.random.rand(len(s.atoms)*3) * u.nanometers
        s.positions = new_crds2
        np.testing.assert_allclose(new_crds2*10,
                s.coordinates.reshape(new_crds2.shape))
        def bad():
            s.positions = [1, 2, 3]
        self.assertRaises(ValueError, bad)
        # Try deleting the coordinates for the first atom
        del s[0].xx, s[0].xy, s[0].xz
        self.assertIs(s.coordinates, None)
        # Make sure it doesn't affect the rest
        for atom in s.view[1:].atoms:
            self.assertTrue(hasattr(atom, 'xx'))
            self.assertTrue(hasattr(atom, 'xy'))
            self.assertTrue(hasattr(atom, 'xz'))
        # Now make sure that setting coordinates to an empty iterable wipes out
        # all xx/y/z attributes
        s.coordinates = []
        self.assertIs(s.coordinates, None)
        # Restore coordinates
        s.coordinates = np.random.rand(3, len(s.atoms), 3)
        del s[0].xx
        self.assertIs(s.get_coordinates(), None)
        s.coordinates = xyz = np.random.rand(3, len(s.atoms), 3)
        del s.atoms[-1]
        np.testing.assert_equal(s.get_coordinates().flatten(), xyz[0,:-1,:].flatten())
        s.coordinates = np.random.rand(3, len(s.atoms), 3)
        s[0].xx = 10
        self.assertEqual(s.get_coordinates().shape, (1, len(s.atoms), 3))
        s.coordinates = xyz = np.random.rand(3, len(s.atoms), 3)
        s[0].xx = 10
        np.testing.assert_equal(s.get_coordinates(0)[1:,:], xyz[0,1:,:])
        s.coordinates = xyz = np.random.rand(3, len(s.atoms), 3)
        s[0].xx = 10
        self.assertRaises(IndexError, lambda: s.get_coordinates(1))

    def test_velocities(self):
        """ Tests coordinate handling in Structure """
        s = create_random_structure(parametrized=False)
        self.assertIs(s.velocities, None)
        x = np.random.rand(len(s.atoms), 3)
        s.velocities = x * u.nanometers/u.picosecond
        np.testing.assert_allclose(x*10, s.velocities)

    def test_coordinate_set_to_empty_list(self):
        """ Tests behavior of setting coordinates to an empty iterable """
        s = create_random_structure(parametrized=True)
        xyz = np.random.random((len(s.atoms), 3))
        s.coordinates = xyz
        # Make sure that the x, y, and z attributes of atoms are equal to the
        # given coordinates
        for atom, pos in zip(s.atoms, xyz):
            self.assertEqual(atom.xx, pos[0])
            self.assertEqual(atom.xy, pos[1])
            self.assertEqual(atom.xz, pos[2])
        # Now set coordinates to an empty list
        s.coordinates = []
        self.assertIs(s.coordinates, None)
        for atom in s.atoms:
            self.assertFalse(hasattr(atom, 'xx'))
            self.assertFalse(hasattr(atom, 'xy'))
            self.assertFalse(hasattr(atom, 'xz'))

    def test_strip(self):
        """ Tests the Structure.strip method """
        s = create_random_structure(parametrized=True)
        nres = len(s.residues)
        natom_per_res = [len(res) for res in s.residues]
        s.strip(':1-5')
        self.assertEqual(len(s.atoms), sum(natom_per_res) - sum(natom_per_res[:5]))
        # Bad usage of strip
        mask = pmd.amber.AmberMask(create_random_structure(parametrized=True), ':1')
        self.assertRaises(TypeError, lambda: s.strip(mask))
        self.assertRaises(TypeError, lambda: s.strip(10))

    def test_strip_array(self):
        """ Tests Structure.strip with a mask array """
        s = create_random_structure(parametrized=True)
        natom = len(s.atoms)
        maskarray = [random.choice([0, 1]) for i in range(natom)]
        s.strip(maskarray)
        self.assertEqual(len(s.atoms), natom-sum(maskarray))
        # Error checking
        self.assertRaises(ValueError, lambda: s.strip([1, 0]))

    def test_strip_with_coords(self):
        """ Tests the Structure.strip method when it has coordinates """
        s = create_random_structure(parametrized=True)
        coords = np.random.rand(10, len(s.atoms), 3)
        s.coordinates = coords
        nres = len(s.residues)
        natom_per_res = [len(res) for res in s.residues]
        s.strip(':1-5')
        self.assertEqual(len(s.atoms), sum(natom_per_res) - sum(natom_per_res[:5]))
        self.assertEqual(s.get_coordinates().shape, (10, len(s.atoms), 3))
        # No longer equal -- it's truncated
        self.assertNotEqual(coords.shape, (10, len(s.atoms), 3))
        # Check that the coordinates left are the same as the original
        # coordinates corresponding to the ones that did *not* get stripped
        n = sum(natom_per_res[:5])
        self.assertTrue((coords[:,n:,:] == s.get_coordinates()).all())

    def test_helpers(self):
        """ Test private helper functions in parmed/structure.py """
        ang, deg = u.angstroms, u.degrees
        strip_units = structure._strip_box_units
        self.assertEqual(strip_units([1, 2, 3]), [1, 2, 3])
        self.assertEqual(strip_units([1*ang, 2*ang, 3*ang]), [1, 2, 3])
        self.assertEqual(strip_units([1*ang, 90*deg]), [1, 90])
        self.assertEqual(strip_units([[1*ang, 90*deg], [2*ang, 109*deg]]),
                         [[1, 90], [2, 109]])
        self.assertRaises(TypeError, lambda: strip_units(['this', 'is', 'bad']))

    def test_radii_assignment(self):
        """ Test radius assignment in Structure """
        parm = pmd.load_file(get_fn('trx.prmtop'))
        criteria_satisfied = [0, 0, 0, 0, 0, 0]
        for atom in parm.atoms:
            if atom.atomic_number == 1:
                oa = atom.bond_partners[0]
                an, mass = oa.atomic_number, oa.mass
                oa.atomic_number = 8
                oa.mass = pmd.periodic_table.Mass['O']
                self.assertEqual(structure._mbondi(atom), 0.8)
                oa.atomic_number = 2
                oa.mass = pmd.periodic_table.Mass['He']
                self.assertEqual(structure._mbondi(atom), 1.2)
                oa.atomic_number = an
                oa.mass = mass
                criteria_satisfied[0] = 1
            if atom.name.startswith('OD') and atom.residue.name == 'ASP':
                self.assertEqual(structure._mbondi3(atom), 1.4)
                criteria_satisfied[1] = 1
            if atom.name.startswith('OXT'):
                self.assertEqual(structure._mbondi3(atom), 1.4)
                criteria_satisfied[2] = 1
            if atom.name.startswith('OE') and atom.residue.name == 'GLU':
                self.assertEqual(structure._mbondi3(atom), 1.4)
                criteria_satisfied[3] = 1
            if atom.name.startswith('HH') and atom.residue.name == 'ARG':
                self.assertEqual(structure._mbondi3(atom), 1.17)
                criteria_satisfied[4] = 1
            if atom.name.startswith('HE') and atom.residue.name == 'ARG':
                self.assertEqual(structure._mbondi3(atom), 1.17)
                criteria_satisfied[5] = 1
        # Make sure we ran all checks
        self.assertEqual(set(criteria_satisfied), {1})

    @unittest.skipUnless(has_openmm, 'Cannot test without OpenMM')
    def test_screen_assignment(self):
        """ Testing screen parameter assignment """
        self.assertEqual(structure._gb_rad_screen(Atom(atomic_number=9), app.GBn2)[1], 0.88)
        self.assertEqual(structure._gb_rad_screen(Atom(atomic_number=15), app.GBn2)[1], 0.86)
        self.assertEqual(structure._gb_rad_screen(Atom(atomic_number=16), app.GBn2)[1], 0.96)
        self.assertEqual(structure._gb_rad_screen(Atom(atomic_number=2), app.GBn2)[1], 0.8)

    @unittest.skipUnless(has_openmm, 'Cannot test without OpenMM')
    def test_omm_topology(self):
        """ Tests the creation of the OpenMM Topology instance """
        # Make sure an empty structure returns an empty topology
        self.assertEqual(len(list(pmd.Structure().topology.atoms())), 0)

    def test_join_dihedrals(self):
        """ Tests the Structure.join_dihedrals method """
        s = self.s
        dt1 = DihedralType(10, 2, 0)
        dt2 = DihedralType(12, 1, 180)
        dt3 = DihedralType(13, 3, 0)
        s.dihedrals.append(Dihedral(s[0], s[1], s[2], s[3], type=dt1))
        s.dihedrals.append(Dihedral(s[3], s[2], s[1], s[0], type=dt2))
        s.dihedrals.append(Dihedral(s[0], s[1], s[2], s[3], type=dt3))
        s.dihedral_types.extend([dt1, dt2, dt3])
        s.dihedral_types.claim()
        # Now join dihedrals
        s.join_dihedrals()
        self.assertEqual(len(s.dihedrals), 1)
        self.assertEqual(len(s.dihedral_types), 1)
        self.assertIsInstance(s.dihedral_types[0], pmd.DihedralTypeList)
        self.assertEqual(len(s[0].dihedral_partners), 3)
        self.assertEqual(len(s[0].dihedrals), 1)

    def test_prune_empty(self):
        """ Tests the prune_empty_terms function """
        s = create_random_structure(parametrized=True)
        # Bonds
        nterms = len(s.bonds)
        s.bonds[-1].delete()
        s.prune_empty_terms()
        self.assertEqual(len(s.bonds), nterms-1)
        # Angles
        nterms = len(s.angles)
        s.angles[-1].delete()
        s.prune_empty_terms()
        self.assertEqual(len(s.angles), nterms-1)
        # Dihedrals
        nterms = len(s.dihedrals)
        s.dihedrals[-1].delete()
        s.prune_empty_terms()
        self.assertEqual(len(s.dihedrals), nterms-1)
        # R-B torsions
        nterms = len(s.rb_torsions)
        s.rb_torsions[-1].delete()
        s.prune_empty_terms()
        self.assertEqual(len(s.rb_torsions), nterms-1)
        # Urey-Bradleys
        nterms = len(s.urey_bradleys)
        s.urey_bradleys[-1].delete()
        s.prune_empty_terms()
        self.assertEqual(len(s.urey_bradleys), nterms-1)
        # Impropers
        nterms = len(s.impropers)
        s.impropers[-1].delete()
        s.prune_empty_terms()
        self.assertEqual(len(s.impropers), nterms-1)
        # CMAPs
        nterms = len(s.cmaps)
        s.cmaps[-1].delete()
        s.prune_empty_terms()
        self.assertEqual(len(s.cmaps), nterms-1)
        # Trigonal angles
        nterms = len(s.trigonal_angles)
        s.trigonal_angles[-1].atom1 = s.trigonal_angles[-1].atom2 = \
                s.trigonal_angles[-1].atom3 = s.trigonal_angles[-1].atom4 = None
        s.prune_empty_terms()
        self.assertEqual(len(s.trigonal_angles), nterms-1)
        # OOP bends
        nterms = len(s.out_of_plane_bends)
        s.out_of_plane_bends[-1].atom1 = s.out_of_plane_bends[-1].atom2 = \
                s.out_of_plane_bends[-1].atom3 = s.out_of_plane_bends[-1].atom4 = None
        s.prune_empty_terms()
        self.assertEqual(len(s.out_of_plane_bends), nterms-1)
        # Pi-torsions
        nterms = len(s.pi_torsions)
        s.pi_torsions[-1].atom1 = s.pi_torsions[-1].atom2 = \
                s.pi_torsions[-1].atom3 = s.pi_torsions[-1].atom4 = \
                s.pi_torsions[-1].atom5 = s.pi_torsions[-1].atom6 = None
        s.prune_empty_terms()
        self.assertEqual(len(s.pi_torsions), nterms-1)
        # Stretch-bends
        nterms = len(s.stretch_bends)
        s.stretch_bends[-1].atom1 = s.stretch_bends[-1].atom2 = \
                s.stretch_bends[-1].atom3 = None
        s.prune_empty_terms()
        self.assertEqual(len(s.stretch_bends), nterms-1)
        # Torsion-torsions
        nterms = len(s.torsion_torsions)
        s.torsion_torsions[-1].delete()
        s.prune_empty_terms()
        self.assertEqual(len(s.torsion_torsions), nterms-1)
        # Chiral frames
        nterms = len(s.chiral_frames)
        s.chiral_frames[-1].atom1 = s.chiral_frames[-1].atom2 = None
        s.prune_empty_terms()
        self.assertEqual(len(s.chiral_frames), nterms-1)
        # Adjusts
        nterms = len(s.adjusts)
        s.adjusts[-1].atom1 = s.adjusts[-1].atom2 = None
        s.prune_empty_terms()
        self.assertEqual(len(s.adjusts), nterms-1)

class TestStructureAdd(FileIOTestCase):
    """ Tests the addition property of a System """

    def _check_mult(self, s, s1, multfac):
        self.assertIsNot(s, s1)
        # Make sure that s is really the sum of s1 and s
        self.assertEqual(len(s.atoms), len(s1.atoms) * multfac)
        self.assertEqual(len(s.residues), len(s1.residues) * multfac)

        def cmp_atoms(a1, a2):
            self.assertIsNot(a1, a2)
            self.assertEqual(a1.name, a2.name)
            self.assertEqual(a1.type, a2.type)
            self.assertEqual(a1.atom_type, a2.atom_type)
            self.assertEqual(a1.type, a2.type)
            self.assertEqual(a1.atom_type, a2.atom_type)
            self.assertEqual(a1.mass, a2.mass)
            self.assertEqual(a1.charge, a2.charge)
            self.assertEqual(a1.atomic_number, a2.atomic_number)
            self.assertEqual(a1.solvent_radius, a2.solvent_radius)
            self.assertEqual(a1.screen, a2.screen)
            self.assertEqual(a1.residue.name, a2.residue.name)
            self.assertEqual(a1.residue.insertion_code, a2.residue.insertion_code)
            self.assertEqual(len(a1.bond_partners), len(a2.bond_partners))
            self.assertEqual(len(a1.angle_partners), len(a2.angle_partners))
            self.assertEqual(len(a1.dihedral_partners), len(a2.dihedral_partners))
            self.assertEqual(len(a1.bonds), len(a2.bonds))
            self.assertEqual(len(a1.angles), len(a2.angles))
            self.assertEqual(len(a1.dihedrals), len(a2.dihedrals))
            self.assertEqual(len(a1.impropers), len(a2.impropers))

        for a1, a2 in zip(s.atoms, s1.atoms * multfac):
            cmp_atoms(a1, a2)
        for r1, r2 in zip(s.residues, s1.residues * multfac):
            self.assertEqual(len(r1), len(r2))
            self.assertEqual(r1.name, r2.name)
            self.assertEqual(r1.chain, r2.chain)
            self.assertEqual(r1.insertion_code, r2.insertion_code)
        self.assertEqual(len(s.bonds), len(s1.bonds)*multfac)
        self.assertEqual(len(s.angles), len(s1.angles)*multfac)
        self.assertEqual(len(s.dihedrals), len(s1.dihedrals)*multfac)
        self.assertEqual(len(s.urey_bradleys), len(s1.urey_bradleys)*multfac)
        self.assertEqual(len(s.impropers), len(s1.impropers)*multfac)
        self.assertEqual(len(s.rb_torsions), len(s1.rb_torsions)*multfac)
        self.assertEqual(len(s.cmaps), len(s1.cmaps)*multfac)
        self.assertEqual(len(s.stretch_bends), len(s1.stretch_bends)*multfac)
        self.assertEqual(len(s.trigonal_angles), len(s1.trigonal_angles)*multfac)
        self.assertEqual(len(s.out_of_plane_bends), len(s1.out_of_plane_bends)*multfac)
        self.assertEqual(len(s.torsion_torsions), len(s1.torsion_torsions)*multfac)
        self.assertEqual(len(s.acceptors), len(s1.acceptors)*multfac)
        self.assertEqual(len(s.donors), len(s1.donors)*multfac)
        self.assertEqual(len(s.pi_torsions), len(s1.pi_torsions)*multfac)
        self.assertEqual(len(s.chiral_frames), len(s1.chiral_frames)*multfac)
        self.assertEqual(len(s.multipole_frames), len(s1.multipole_frames)*multfac)
        self.assertEqual(len(s.groups), len(s1.groups)*multfac)
        self.assertEqual(len(s.adjusts), len(s1.adjusts)*multfac)
        # Check types
        self.assertEqual(len(s.bond_types), len(s1.bond_types))
        self.assertEqual(len(s.angle_types), len(s1.angle_types))
        self.assertEqual(len(s.dihedral_types), len(s1.dihedral_types))
        self.assertEqual(len(s.urey_bradley_types), len(s1.urey_bradley_types))
        self.assertEqual(len(s.improper_types), len(s1.improper_types))
        self.assertEqual(len(s.rb_torsion_types), len(s1.rb_torsion_types))
        self.assertEqual(len(s.cmap_types), len(s1.cmap_types))
        self.assertEqual(len(s.stretch_bend_types), len(s1.stretch_bend_types))
        self.assertEqual(len(s.trigonal_angle_types), len(s1.trigonal_angle_types))
        self.assertEqual(len(s.out_of_plane_bend_types), len(s1.out_of_plane_bend_types))
        self.assertEqual(len(s.torsion_torsion_types), len(s1.torsion_torsion_types))
        self.assertEqual(len(s.pi_torsion_types), len(s1.pi_torsion_types))
        self.assertEqual(len(s.adjust_types), len(s1.adjust_types))
        # Check all valence terms
        def chk_valence(val1, val2):
            self.assertIs(type(val1[0]), type(val2[0]))
            self.assertEqual(len(val1), len(val2))
            attrs = [attr for attr in dir(val1[0]) if attr.startswith('atom')]
            for v1, v2 in zip(val1, val2):
                at1 = [getattr(v1, attr) for attr in attrs]
                at2 = [getattr(v2, attr) for attr in attrs]
                self.assertIsNot(v1, v2)
                for a1, a2 in zip(at1, at2):
                    cmp_atoms(a1, a2)
                if hasattr(v1, 'type'):
                    self.assertEqual(v1.type, v2.type)
                    if not isinstance(v1.type, int):
                        if v1.type is None:
                            self.assertIs(v2.type, None)
                        else:
                            self.assertIsNot(v1.type, v2.type)
        chk_valence(s.bonds, s1.bonds*multfac)
        chk_valence(s.angles, s1.angles*multfac)
        chk_valence(s.dihedrals, s1.dihedrals*multfac)
        chk_valence(s.rb_torsions, s1.rb_torsions*multfac)
        chk_valence(s.urey_bradleys, s1.urey_bradleys*multfac)
        chk_valence(s.impropers, s1.impropers*multfac)
        chk_valence(s.cmaps, s1.cmaps*multfac)
        chk_valence(s.trigonal_angles, s1.trigonal_angles*multfac)
        chk_valence(s.out_of_plane_bends, s1.out_of_plane_bends*multfac)
        chk_valence(s.pi_torsions, s1.pi_torsions*multfac)
        chk_valence(s.torsion_torsions, s1.torsion_torsions*multfac)
        chk_valence(s.stretch_bends, s1.stretch_bends*multfac)
        chk_valence(s.chiral_frames, s1.chiral_frames*multfac)
        chk_valence(s.multipole_frames, s1.multipole_frames*multfac)
        chk_valence(s.donors, s1.donors*multfac)
        chk_valence(s.acceptors, s1.acceptors*multfac)
        chk_valence(s.groups, s1.groups*multfac)

    def _check_sum(self, s, s1, s2):
        self.assertIsNot(s, s1)
        self.assertIsNot(s, s2)
        # Make sure that s is really the sum of s1 and s2
        self.assertEqual(len(s.atoms), len(s1.atoms) + len(s2.atoms))
        self.assertEqual(len(s.residues), len(s1.residues) + len(s2.residues))

        def cmp_atoms(a1, a2):
            self.assertIsNot(a1, a2)
            self.assertEqual(a1.name, a2.name)
            if isinstance(a2.type, str) and f"{a2.type}D" == a1.type:
                a2type = copy(a2.atom_type)
                a2type.name += "D"
                self.assertEqual(a1.atom_type, a2type)
            else:
                self.assertEqual(a1.type, a2.type)
                self.assertEqual(a1.atom_type, a2.atom_type)
            self.assertAlmostEqual(a1.mass, a2.mass)
            self.assertAlmostEqual(a1.charge, a2.charge)
            self.assertEqual(a1.atomic_number, a2.atomic_number)
            self.assertAlmostEqual(a1.solvent_radius, a2.solvent_radius)
            self.assertAlmostEqual(a1.screen, a2.screen)
            self.assertEqual(a1.residue.name, a2.residue.name)
            self.assertEqual(a1.residue.insertion_code, a2.residue.insertion_code)
            self.assertEqual(len(a1.bond_partners), len(a2.bond_partners))
            self.assertEqual(len(a1.angle_partners), len(a2.angle_partners))
            self.assertEqual(len(a1.dihedral_partners), len(a2.dihedral_partners))
            self.assertEqual(len(a1.bonds), len(a2.bonds))
            self.assertEqual(len(a1.angles), len(a2.angles))
            self.assertEqual(len(a1.dihedrals), len(a2.dihedrals))
            self.assertEqual(len(a1.impropers), len(a2.impropers))

        for a1, a2 in zip(s.atoms, s1.atoms + s2.atoms):
            cmp_atoms(a1, a2)
        for r1, r2 in zip(s.residues, s1.residues + s2.residues):
            self.assertEqual(len(r1), len(r2))
            self.assertEqual(r1.name, r2.name)
            self.assertEqual(r1.chain, r2.chain)
            self.assertEqual(r1.insertion_code, r2.insertion_code)
        self.assertEqual(len(s.bonds), len(s1.bonds)+len(s2.bonds))
        self.assertEqual(len(s.angles), len(s1.angles)+len(s2.angles))
        self.assertEqual(len(s.dihedrals), len(s1.dihedrals)+len(s2.dihedrals))
        self.assertEqual(len(s.urey_bradleys), len(s1.urey_bradleys)+len(s2.urey_bradleys))
        self.assertEqual(len(s.impropers), len(s1.impropers)+len(s2.impropers))
        self.assertEqual(len(s.rb_torsions), len(s1.rb_torsions)+len(s2.rb_torsions))
        self.assertEqual(len(s.cmaps), len(s1.cmaps)+len(s2.cmaps))
        self.assertEqual(len(s.stretch_bends), len(s1.stretch_bends)+len(s2.stretch_bends))
        self.assertEqual(len(s.trigonal_angles), len(s1.trigonal_angles)+len(s2.trigonal_angles))
        self.assertEqual(len(s.out_of_plane_bends), len(s1.out_of_plane_bends)+len(s2.out_of_plane_bends))
        self.assertEqual(len(s.torsion_torsions), len(s1.torsion_torsions)+len(s2.torsion_torsions))
        self.assertEqual(len(s.acceptors), len(s1.acceptors)+len(s2.acceptors))
        self.assertEqual(len(s.donors), len(s1.donors)+len(s2.donors))
        self.assertEqual(len(s.pi_torsions), len(s1.pi_torsions)+len(s2.pi_torsions))
        self.assertEqual(len(s.chiral_frames), len(s1.chiral_frames)+len(s2.chiral_frames))
        self.assertEqual(len(s.multipole_frames), len(s1.multipole_frames)+len(s2.multipole_frames))
        self.assertEqual(len(s.groups), len(s1.groups)+len(s2.groups))
        self.assertEqual(len(s.adjusts), len(s1.adjusts)+len(s2.adjusts))
        # Check types
        self.assertEqual(len(s.bond_types), len(s1.bond_types) + len(s2.bond_types))
        self.assertEqual(len(s.angle_types), len(s1.angle_types)+len(s2.angle_types))
        self.assertEqual(len(s.dihedral_types), len(s1.dihedral_types)+len(s2.dihedral_types))
        self.assertEqual(len(s.urey_bradley_types), len(s1.urey_bradley_types)+len(s2.urey_bradley_types))
        self.assertEqual(len(s.improper_types), len(s1.improper_types)+len(s2.improper_types))
        self.assertEqual(len(s.rb_torsion_types), len(s1.rb_torsion_types)+len(s2.rb_torsion_types))
        self.assertEqual(len(s.cmap_types), len(s1.cmap_types)+len(s2.cmap_types))
        self.assertEqual(len(s.stretch_bend_types), len(s1.stretch_bend_types)+len(s2.stretch_bend_types))
        self.assertEqual(len(s.trigonal_angle_types), len(s1.trigonal_angle_types)+len(s2.trigonal_angle_types))
        self.assertEqual(len(s.out_of_plane_bend_types), len(s1.out_of_plane_bend_types)+len(s2.out_of_plane_bend_types))
        self.assertEqual(len(s.torsion_torsion_types), len(s1.torsion_torsion_types)+len(s2.torsion_torsion_types))
        self.assertEqual(len(s.pi_torsion_types), len(s1.pi_torsion_types)+len(s2.pi_torsion_types))
        #self.assertEqual(len(s.adjust_types), len(s1.adjust_types)+len(s2.adjust_types))
        self.assertEqual(len(set(s.adjust_types)), len(set(s1.adjust_types) | set(s2.adjust_types)))
        # Check all valence terms
        def chk_valence(val1, val2):
            if not val1 or not val2:
                self.assertFalse(val1)
                self.assertFalse(val2)
                return
            self.assertIs(type(val1[0]), type(val2[0]))
            self.assertEqual(len(val1), len(val2))
            attrs = [attr for attr in dir(val1[0]) if attr.startswith('atom')]
            for v1, v2 in zip(val1, val2):
                at1 = [getattr(v1, attr) for attr in attrs]
                at2 = [getattr(v2, attr) for attr in attrs]
                self.assertIsNot(v1, v2)
                for a1, a2 in zip(at1, at2):
                    cmp_atoms(a1, a2)
                if hasattr(v1, 'type'):
                    self.assertEqual(v1.type, v2.type)
                    if not isinstance(v1.type, int):
                        self.assertIsNot(v1.type, v2.type)
        chk_valence(s.bonds, s1.bonds+s2.bonds)
        chk_valence(s.angles, s1.angles+s2.angles)
        chk_valence(s.dihedrals, s1.dihedrals+s2.dihedrals)
        chk_valence(s.rb_torsions, s1.rb_torsions + s2.rb_torsions)
        chk_valence(s.urey_bradleys, s1.urey_bradleys+s2.urey_bradleys)
        chk_valence(s.impropers, s1.impropers+s2.impropers)
        chk_valence(s.cmaps, s1.cmaps+s2.cmaps)
        chk_valence(s.trigonal_angles, s1.trigonal_angles+s2.trigonal_angles)
        chk_valence(s.out_of_plane_bends, s1.out_of_plane_bends+s2.out_of_plane_bends)
        chk_valence(s.pi_torsions, s1.pi_torsions+s2.pi_torsions)
        chk_valence(s.torsion_torsions, s1.torsion_torsions+s2.torsion_torsions)
        chk_valence(s.stretch_bends, s1.stretch_bends+s2.stretch_bends)
        chk_valence(s.chiral_frames, s1.chiral_frames+s2.chiral_frames)
        chk_valence(s.multipole_frames, s1.multipole_frames+s2.multipole_frames)
        chk_valence(s.donors, s1.donors+s2.donors)
        chk_valence(s.acceptors, s1.acceptors+s2.acceptors)
        chk_valence(s.groups, s1.groups+s2.groups)

    def test_add_parametrized(self):
        """ Tests addition of two parametrized Structure instances """
        s1 = create_random_structure(parametrized=True)
        s2 = create_random_structure(parametrized=True)
        s1.coordinates = np.random.random((4, len(s1.atoms), 3))
        s2.coordinates = np.random.random((2, len(s2.atoms), 3))
        s1.box = [[10, 10, 10, 90, 90, 90], [11, 11, 11, 90, 90, 90]]
        s2.box = [[20, 20, 20, 90, 90, 90], [21, 21, 21, 90, 90, 90]]
        self.assertTrue(bool(s1.bond_types))
        self.assertTrue(bool(s2.bond_types))
        s = s1 + s2
        self._check_sum(s, s1, s2)
        self.assertEqual(s.get_coordinates('all').shape, (2, len(s.atoms), 3))

    def test_add_issue_1197(self):
        """ Tests addition of two Structures with overlapping atom type names """
        with open(self.get_fn("thf.pkl"), "rb") as f:
            s1 = pickle.load(f)
            s1.residues[0].name = "THF"
            # Add new attributes since this is a saved pickle file that predates adding these new attributes
            for atom in s1.atoms:
                atom.formal_charge = atom.hybridization = atom.aromatic = None
            for bond in s1.bonds:
                bond.qualitative_type = None
        with open(self.get_fn("ace.pkl"), "rb") as f:
            s2 = pickle.load(f)
            s2.residues[0].name = "ACE"
            for atom in s2.atoms:
                atom.formal_charge = atom.hybridization = atom.aromatic = None
            for bond in s2.bonds:
                bond.qualitative_type = None
        combined_fn = self.get_fn("gmx-co1bined.top", written=True)
        comb = s1 + s2
        comb.save(combined_fn, format="GROMACS")
        s = pmd.load_file(combined_fn)
        s.residues[0].chain = s1.residues[0].chain
        s.residues[1].chain = s2.residues[0].chain

        # The 1-4 charge scaling is handled by the adjust arrays, so go ahead and copy those over to avoid false fails
        for d1, d2 in zip(s.dihedrals, s1.dihedrals + s2.dihedrals):
            d1.type.scee = d2.type.scee
            d1.type.scnb = d2.type.scnb

        # We have some precision loss with writing to GROMACS (but this is necessary to trigger the original bug).
        # So tweak our comparison thresholds for *just* this comparison.
        with mock.patch("parmed.topologyobjects.TINY", 10e-4), mock.patch("parmed.topologyobjects._TINY_DIGITS", 4):
            self._check_sum(s, s1, s2)

    def test_add_to_empty_structure(self):
        """ Tests addition to empty Structure """
        s1 = create_random_structure(parametrized=True)
        s2 = structure.Structure()
        s2 += s1
        self._check_sum(s2, structure.Structure(), s1)

    def test_iadd(self):
        """ Tests in-place addition of two Structure instances """
        s1 = create_random_structure(parametrized=True)
        s2 = create_random_structure(parametrized=True)
        s1cp = copy(s1)
        s1 += s2
        self._check_sum(s1, s1cp, s2)

    def test_copy_bond_order(self):
        """ Tests that copying a structure also copies bond order """
        s1 = create_random_structure(parametrized=True)
        s1.bonds[0].order = 1.25

        s1cp = copy(s1)
        self.assertEqual(s1cp.bonds[0].order, 1.25)

    def test_slice_bond_order(self):
        """ Tests that copying via slice also copies bond order """
        s1 = create_random_structure(parametrized=True)
        s1.bonds[0].order = 1.25

        s1cp = s1[:]
        self.assertEqual(s1cp.bonds[0].order, 1.25)

    def test_add_not_parametrized(self):
        """ Tests addition of two non-parametrized Structure instances """
        s1 = create_random_structure(parametrized=False)
        s2 = create_random_structure(parametrized=False)
        self.assertFalse(bool(s1.bond_types))
        self.assertFalse(bool(s2.bond_types))
        s = s1 + s2
        self.assertIsNot(s, s1)
        self.assertIsNot(s, s2)
        # Make sure that s is really the sum of s1 and s2
        self.assertEqual(len(s.atoms), len(s1.atoms) + len(s2.atoms))
        self.assertEqual(len(s.residues), len(s1.residues) + len(s2.residues))

        def cmp_atoms(a1, a2):
            self.assertIsNot(a1, a2)
            self.assertEqual(a1.name, a2.name)
            self.assertEqual(a1.type, a2.type)
            self.assertEqual(a1.atom_type, a2.atom_type)
            self.assertEqual(a1.mass, a2.mass)
            self.assertEqual(a1.charge, a2.charge)
            self.assertEqual(a1.atomic_number, a2.atomic_number)
            self.assertEqual(a1.solvent_radius, a2.solvent_radius)
            self.assertEqual(a1.screen, a2.screen)
            self.assertEqual(a1.residue.name, a2.residue.name)
            self.assertEqual(a1.residue.insertion_code, a2.residue.insertion_code)
            self.assertEqual(len(a1.bond_partners), len(a2.bond_partners))
            self.assertEqual(len(a1.angle_partners), len(a2.angle_partners))
            self.assertEqual(len(a1.dihedral_partners), len(a2.dihedral_partners))
            self.assertEqual(len(a1.bonds), len(a2.bonds))
            self.assertEqual(len(a1.angles), len(a2.angles))
            self.assertEqual(len(a1.dihedrals), len(a2.dihedrals))
            self.assertEqual(len(a1.impropers), len(a2.impropers))

        for a1, a2 in zip(s.atoms, s1.atoms + s2.atoms):
            cmp_atoms(a1, a2)
        for r1, r2 in zip(s.residues, s1.residues + s2.residues):
            self.assertEqual(len(r1), len(r2))
            self.assertEqual(r1.name, r2.name)
            self.assertEqual(r1.chain, r2.chain)
            self.assertEqual(r1.insertion_code, r2.insertion_code)
        self.assertEqual(len(s.bonds), len(s1.bonds)+len(s2.bonds))
        self.assertEqual(len(s.angles), len(s1.angles)+len(s2.angles))
        self.assertEqual(len(s.dihedrals), len(s1.dihedrals)+len(s2.dihedrals))
        self.assertEqual(len(s.urey_bradleys), len(s1.urey_bradleys)+len(s2.urey_bradleys))
        self.assertEqual(len(s.impropers), len(s1.impropers)+len(s2.impropers))
        self.assertEqual(len(s.rb_torsions), len(s1.rb_torsions)+len(s2.rb_torsions))
        self.assertEqual(len(s.cmaps), len(s1.cmaps)+len(s2.cmaps))
        self.assertEqual(len(s.stretch_bends), len(s1.stretch_bends)+len(s2.stretch_bends))
        self.assertEqual(len(s.trigonal_angles), len(s1.trigonal_angles)+len(s2.trigonal_angles))
        self.assertEqual(len(s.out_of_plane_bends), len(s1.out_of_plane_bends)+len(s2.out_of_plane_bends))
        self.assertEqual(len(s.torsion_torsions), len(s1.torsion_torsions)+len(s2.torsion_torsions))
        self.assertEqual(len(s.acceptors), len(s1.acceptors)+len(s2.acceptors))
        self.assertEqual(len(s.donors), len(s1.donors)+len(s2.donors))
        self.assertEqual(len(s.pi_torsions), len(s1.pi_torsions)+len(s2.pi_torsions))
        self.assertEqual(len(s.chiral_frames), len(s1.chiral_frames)+len(s2.chiral_frames))
        self.assertEqual(len(s.multipole_frames), len(s1.multipole_frames)+len(s2.multipole_frames))
        self.assertEqual(len(s.groups), len(s1.groups)+len(s2.groups))
        self.assertEqual(len(s.adjusts), len(s1.adjusts)+len(s2.adjusts))
        self.assertEqual(len(s.adjust_types), len(s1.adjust_types)+len(s2.adjust_types))
        # Check all valence terms
        def chk_valence(val1, val2):
            self.assertIs(type(val1[0]), type(val2[0]))
            self.assertEqual(len(val1), len(val2))
            attrs = [attr for attr in dir(val1[0]) if attr.startswith('atom')]
            for v1, v2 in zip(val1, val2):
                at1 = [getattr(v1, attr) for attr in attrs]
                at2 = [getattr(v2, attr) for attr in attrs]
                self.assertIsNot(v1, v2)
                for a1, a2 in zip(at1, at2):
                    cmp_atoms(a1, a2)
        chk_valence(s.bonds, s1.bonds+s2.bonds)
        chk_valence(s.angles, s1.angles+s2.angles)
        chk_valence(s.dihedrals, s1.dihedrals+s2.dihedrals)
        chk_valence(s.rb_torsions, s1.rb_torsions+s2.rb_torsions)
        chk_valence(s.urey_bradleys, s1.urey_bradleys+s2.urey_bradleys)
        chk_valence(s.impropers, s1.impropers+s2.impropers)
        chk_valence(s.cmaps, s1.cmaps+s2.cmaps)
        chk_valence(s.trigonal_angles, s1.trigonal_angles+s2.trigonal_angles)
        chk_valence(s.out_of_plane_bends, s1.out_of_plane_bends+s2.out_of_plane_bends)
        chk_valence(s.pi_torsions, s1.pi_torsions+s2.pi_torsions)
        chk_valence(s.torsion_torsions, s1.torsion_torsions+s2.torsion_torsions)
        chk_valence(s.stretch_bends, s1.stretch_bends+s2.stretch_bends)
        chk_valence(s.chiral_frames, s1.chiral_frames+s2.chiral_frames)
        chk_valence(s.multipole_frames, s1.multipole_frames+s2.multipole_frames)
        chk_valence(s.donors, s1.donors+s2.donors)
        chk_valence(s.acceptors, s1.acceptors+s2.acceptors)
        chk_valence(s.groups, s1.groups+s2.groups)

    def test_add_no_valence(self):
        """ Tests addition of two minimal Structure instances """
        s1 = create_random_structure(parametrized=False, novalence=True)
        s2 = create_random_structure(parametrized=False, novalence=True)
        s = s1 + s2
        self.assertIsNot(s, s1)
        self.assertIsNot(s, s2)
        # Make sure that s is really the sum of s1 and s2
        self.assertEqual(len(s.atoms), len(s1.atoms) + len(s2.atoms))
        self.assertEqual(len(s.residues), len(s1.residues) + len(s2.residues))

        def cmp_atoms(a1, a2):
            self.assertIsNot(a1, a2)
            self.assertEqual(a1.name, a2.name)
            self.assertEqual(a1.type, a2.type)
            self.assertEqual(a1.atom_type, a2.atom_type)
            self.assertEqual(a1.mass, a2.mass)
            self.assertEqual(a1.charge, a2.charge)
            self.assertEqual(a1.atomic_number, a2.atomic_number)
            self.assertEqual(a1.solvent_radius, a2.solvent_radius)
            self.assertEqual(a1.screen, a2.screen)
            self.assertEqual(a1.residue.name, a2.residue.name)
            self.assertEqual(a1.residue.insertion_code, a2.residue.insertion_code)
            self.assertEqual(len(a1.bond_partners), len(a2.bond_partners))
            self.assertEqual(len(a1.angle_partners), len(a2.angle_partners))
            self.assertEqual(len(a1.dihedral_partners), len(a2.dihedral_partners))
            self.assertEqual(len(a1.bonds), len(a2.bonds))
            self.assertEqual(len(a1.angles), len(a2.angles))
            self.assertEqual(len(a1.dihedrals), len(a2.dihedrals))
            self.assertEqual(len(a1.impropers), len(a2.impropers))

        for a1, a2 in zip(s.atoms, s1.atoms + s2.atoms):
            cmp_atoms(a1, a2)
        for r1, r2 in zip(s.residues, s1.residues + s2.residues):
            self.assertEqual(len(r1), len(r2))
            self.assertEqual(r1.name, r2.name)
            self.assertEqual(r1.chain, r2.chain)
            self.assertEqual(r1.insertion_code, r2.insertion_code)

    def test_multiply_parametrized(self):
        """ Tests replicating a parametrized Structure instance """
        s1 = create_random_structure(parametrized=True)
        multfac = random.randint(2, 5)
        s2 = s1 * multfac
        self.assertIsNot(s1, s2)
        self.assertEqual(len(s2.atoms), len(s1.atoms) * multfac)
        self._check_mult(s2, s1, multfac)
        # Check commutativity
        s3 = multfac * s1
        self.assertIsNot(s1, s3)
        self.assertEqual(len(s3.atoms), len(s1.atoms) * multfac)
        self._check_mult(s3, s1, multfac)
        # Check when coordinates exist
        crd = np.random.rand(5, len(s1.atoms), 3)
        s1.coordinates = crd
        np.testing.assert_equal((s1*multfac).get_coordinates(),
                                np.hstack([crd for i in range(multfac)]))

    def test_multiply_not_parametrized(self):
        """ Tests replicating a non-parametrized Structure instance """
        s1 = create_random_structure(parametrized=False)
        multfac = random.randint(2, 5)
        s2 = s1 * multfac
        self.assertIsNot(s1, s2)
        self.assertEqual(len(s2.atoms), len(s1.atoms) * multfac)
        self._check_mult(s2, s1, multfac)
        # Check commutativity
        s3 = multfac * s1
        self.assertIsNot(s1, s3)
        self.assertEqual(len(s3.atoms), len(s1.atoms) * multfac)
        self._check_mult(s3, s1, multfac)

    def test_mult_no_valence(self):
        """ Tests addition of two minimal Structure instances """
        s1 = create_random_structure(parametrized=False, novalence=True)
        multfac = random.randint(2, 5)
        s = s1 * multfac
        self.assertIsNot(s1, s)
        self.assertEqual(len(s.atoms), len(s1.atoms) * multfac)
        self.assertIsNot(s, s1)
        # Make sure that s is really the sum of s1 and s
        self.assertEqual(len(s.atoms), len(s1.atoms) * multfac)
        self.assertEqual(len(s.residues), len(s1.residues) * multfac)

        def cmp_atoms(a1, a2):
            self.assertIsNot(a1, a2)
            self.assertEqual(a1.name, a2.name)
            self.assertEqual(a1.type, a2.type)
            self.assertEqual(a1.atom_type, a2.atom_type)
            self.assertEqual(a1.mass, a2.mass)
            self.assertEqual(a1.charge, a2.charge)
            self.assertEqual(a1.atomic_number, a2.atomic_number)
            self.assertEqual(a1.solvent_radius, a2.solvent_radius)
            self.assertEqual(a1.screen, a2.screen)
            self.assertEqual(a1.residue.name, a2.residue.name)
            self.assertEqual(a1.residue.insertion_code, a2.residue.insertion_code)
            self.assertEqual(len(a1.bond_partners), len(a2.bond_partners))
            self.assertEqual(len(a1.angle_partners), len(a2.angle_partners))
            self.assertEqual(len(a1.dihedral_partners), len(a2.dihedral_partners))
            self.assertEqual(len(a1.bonds), len(a2.bonds))
            self.assertEqual(len(a1.angles), len(a2.angles))
            self.assertEqual(len(a1.dihedrals), len(a2.dihedrals))
            self.assertEqual(len(a1.impropers), len(a2.impropers))

        for a1, a2 in zip(s.atoms, s1.atoms * multfac):
            cmp_atoms(a1, a2)
        for r1, r2 in zip(s.residues, s1.residues * multfac):
            self.assertEqual(len(r1), len(r2))
            self.assertEqual(r1.name, r2.name)
            self.assertEqual(r1.chain, r2.chain)
            self.assertEqual(r1.insertion_code, r2.insertion_code)

class TestStructureSave(FileIOTestCase):
    """ Tests the universal "save" function in Structure """

    def setUp(self):
        self.sys1 = pmd.load_file(get_fn('ala3_solv.psf'))
        self.sys1.coordinates = pmd.load_file(get_fn('ala3_solv.crd')).coordinates
        self.sys1.box = [3.271195e1, 3.299596e1, 3.300715e1, 90, 90, 90]
        self.sys1.load_parameters(
                pmd.charmm.CharmmParameterSet(
                        get_fn('par_all36_prot.prm'),
                        get_fn('toppar_water_ions.str')
                )
        )
        self.sys2 = pmd.load_file(get_fn('trx.prmtop'), get_fn('trx.inpcrd'))
        self.sys3 = pmd.load_file(get_fn(os.path.join('01.1water', 'topol.top')),
                                  xyz=get_fn(os.path.join('01.1water', 'conf.gro')))
        self.sys4 = pmd.load_file(get_fn('ala_ala_ala.psf'))
        self.sys4.coordinates = np.random.rand(len(self.sys4.atoms), 3)
        super(TestStructureSave, self).setUp()

    def test_save_pdb(self):
        """ Test saving various Structure instances as a PDB """
        self.sys1.save(self.get_fn('test.pdb', written=True))
        self.sys2.save(self.get_fn('test2.pdb', written=True))
        self.sys3.save(self.get_fn('test3.pdb', written=True))
        self.sys4.save(self.get_fn('test4.pdb', written=True))
        stringio_file = StringIO()
        self.sys4.save(stringio_file, format='pdb')
        stringio_file.seek(0)
        self.assertTrue(pmd.formats.PDBFile.id_format(stringio_file))
        stringio_file.seek(0)
        x1 = pmd.formats.PDBFile.parse(self.get_fn('test.pdb', written=True))
        x2 = pmd.formats.PDBFile.parse(self.get_fn('test2.pdb', written=True))
        x3 = pmd.formats.PDBFile.parse(self.get_fn('test3.pdb', written=True))
        x4 = pmd.formats.PDBFile.parse(self.get_fn('test4.pdb', written=True))
        x5 = pmd.formats.PDBFile.parse(stringio_file)
        self.assertEqual([a.name for a in self.sys1.atoms], [a.name for a in x1.atoms])
        self.assertEqual([a.name for a in self.sys2.atoms], [a.name for a in x2.atoms])
        self.assertEqual([a.name for a in self.sys3.atoms], [a.name for a in x3.atoms])
        self.assertEqual([a.name for a in self.sys4.atoms], [a.name for a in x4.atoms])
        self.assertEqual([a.name for a in self.sys4.atoms], [a.name for a in x5.atoms])
        # Make sure atom types as integers are preserved
        self.sys4.load_parameters(
                pmd.charmm.CharmmParameterSet(
                    get_fn('top_all22_prot.inp'),
                    get_fn('par_all22_prot.inp')
                )
        )
        for a in self.sys4.atoms:
            a.type = int(a.atom_type)
        self.sys4.save(self.get_fn('test5.pdb', written=True))
        for a in self.sys4.atoms:
            self.assertIsInstance(a.type, int)

        # raise if specifying file-like object and format is None
        stringio_file = StringIO()
        self.assertRaises(RuntimeError, lambda: self.sys4.save(stringio_file))

    def test_save_cif(self):
        """ Test saving various Structure instances as a PDBx/mmCIF """
        self.sys1.save(self.get_fn('test.cif', written=True))
        self.sys2.save(self.get_fn('test2.cif', written=True))
        self.sys3.save(self.get_fn('test3.cif', written=True))
        stringio_file = StringIO()
        self.sys3.save(stringio_file, format='cif')
        stringio_file.seek(0)
        x1 = pmd.formats.CIFFile.parse(self.get_fn('test.cif', written=True))
        x2 = pmd.formats.CIFFile.parse(self.get_fn('test2.cif', written=True))
        x3 = pmd.formats.CIFFile.parse(self.get_fn('test3.cif', written=True))
        x4 = pmd.formats.CIFFile.parse(stringio_file)
        self.assertEqual([a.name for a in self.sys1.atoms], [a.name for a in x1.atoms])
        self.assertEqual([a.name for a in self.sys2.atoms], [a.name for a in x2.atoms])
        self.assertEqual([a.name for a in self.sys3.atoms], [a.name for a in x3.atoms])
        self.assertEqual([a.name for a in self.sys3.atoms], [a.name for a in x4.atoms])
        # Try a gzip and a bzip2 file
        self.sys1.save(self.get_fn('test.cif.bz2', written=True))
        self.sys1.save(self.get_fn('test.cif.gz', written=True))
        # Make sure they're compressed
        bz2.BZ2File(self.get_fn('test.cif.bz2', written=True), 'r').read()
        gzip.open(self.get_fn('test.cif.gz', written=True), 'r').read()
        # Make sure they're the right format
        self.assertTrue(pmd.formats.CIFFile.id_format(self.get_fn('test.cif.gz', written=True)))
        self.assertTrue(pmd.formats.CIFFile.id_format(self.get_fn('test.cif.bz2', written=True)))

    def test_save_mol2(self):
        """ Test saving various Structure instances as Mol2 files """
        self.sys1.save(self.get_fn('test.mol2', written=True))
        self.sys2.save(self.get_fn('test2.mol2', written=True))
        self.sys3.save(self.get_fn('test3.mol2', written=True))
        stringio_file = StringIO()
        self.sys3.save(stringio_file, format='mol2')
        stringio_file.seek(0)
        x1 = pmd.formats.Mol2File.parse(self.get_fn('test.mol2', written=True), structure=True)
        x2 = pmd.formats.Mol2File.parse(self.get_fn('test2.mol2', written=True), structure=True)
        x3 = pmd.formats.Mol2File.parse(self.get_fn('test3.mol2', written=True), structure=True)
        x4 = pmd.formats.Mol2File.parse(stringio_file, structure=True)
        self.assertEqual([a.name for a in self.sys1.atoms], [a.name for a in x1.atoms])
        self.assertEqual([a.name for a in self.sys2.atoms], [a.name for a in x2.atoms])
        self.assertEqual([a.name for a in self.sys3.atoms], [a.name for a in x3.atoms])
        self.assertEqual([a.name for a in self.sys3.atoms], [a.name for a in x4.atoms])
        self.assertEqual(len(self.sys1.bonds), len(x1.bonds))
        self.assertEqual(len(self.sys2.bonds), len(x2.bonds))
        self.assertEqual(len(self.sys3.bonds), len(x3.bonds))

    def test_save_mol3(self):
        """ Test saving various Structure instances as Mol3 files """
        self.sys1.save(self.get_fn('test.mol3', written=True))
        stringio_file = StringIO()
        self.sys1.save(stringio_file, format='mol3')
        stringio_file.seek(0)
        x1 = pmd.formats.Mol2File.parse(self.get_fn('test.mol3', written=True), structure=True)
        x2 = pmd.formats.Mol2File.parse(stringio_file, structure=True)
        self.assertEqual([a.name for a in self.sys1.atoms], [a.name for a in x1.atoms])
        self.assertEqual([a.name for a in self.sys1.atoms], [a.name for a in x2.atoms])
        self.assertEqual(len(self.sys1.bonds), len(x1.bonds))
        with open(self.get_fn('test.mol3', written=True), 'r') as f:
            for line in f:
                if line.startswith('@<TRIPOS>HEADTAIL'):
                    break
            else:
                self.assertFalse(True)

    def test_save_amber_parm(self):
        """ Test saving various Structure instances as Amber prmtop files """
        self.sys1.save(self.get_fn('test.parm7', written=True))
        self.sys2.save(self.get_fn('test2.parm7', written=True))
        self.sys3.save(self.get_fn('test3.parm7', written=True))
        stringio_file = StringIO()
        self.sys3.save(stringio_file, format='amber')
        stringio_file.seek(0)
        x1 = pmd.amber.LoadParm(self.get_fn('test.parm7', written=True))
        x2 = pmd.amber.LoadParm(self.get_fn('test2.parm7', written=True))
        x3 = pmd.amber.LoadParm(self.get_fn('test3.parm7', written=True))
        x4 = pmd.amber.LoadParm(stringio_file)
        self.assertIsInstance(x1, pmd.amber.ChamberParm)
        self.assertIsInstance(x2, pmd.amber.AmberParm)
        self.assertIsInstance(x3, pmd.amber.AmberParm)
        self.assertIsInstance(x4, pmd.amber.AmberParm)
        # Check equivalence of topologies
        self.assertEqual([a.name for a in self.sys1.atoms], [a.name for a in x1.atoms])
        self.assertEqual([a.name for a in self.sys2.atoms], [a.name for a in x2.atoms])
        self.assertEqual([a.name for a in self.sys3.atoms], [a.name for a in x3.atoms])
        self.assertEqual([a.name for a in self.sys3.atoms], [a.name for a in x4.atoms])
        self.assertEqual(len(self.sys1.bonds), len(x1.bonds))
        self.assertEqual(len(self.sys2.bonds), len(x2.bonds))
        self.assertEqual(len(self.sys3.bonds), len(x3.bonds))
        self.assertEqual(len(self.sys1.angles), len(x1.angles))
        self.assertEqual(len(self.sys2.angles), len(x2.angles))
        self.assertEqual(len(self.sys3.angles), len(x3.angles))
        self.assertEqual(sum([len(d.type) for d in self.sys1.dihedrals]), len(x1.dihedrals))
        self.assertEqual(len(self.sys2.dihedrals), len(x2.dihedrals))
        self.assertEqual(sum([len(d.type) for d in self.sys3.dihedrals]), len(x3.dihedrals))
        self.assertEqual(len(self.sys1.impropers), len(x1.impropers))
        self.assertEqual(len(self.sys2.impropers), len(x2.impropers))
        self.assertEqual(len(self.sys3.impropers), len(x3.impropers))
        self.assertEqual(len(self.sys1.cmaps), len(x1.cmaps))
        self.assertEqual(len(self.sys2.cmaps), len(x2.cmaps))
        self.assertEqual(len(self.sys3.cmaps), len(x3.cmaps))
        for a1, a2 in zip(self.sys1.atoms, x1.atoms):
            self.assertAlmostEqual(a1.rmin, a2.rmin)
            self.assertAlmostEqual(abs(a1.epsilon), abs(a2.epsilon))
        for a1, a2 in zip(self.sys2.atoms, x2.atoms):
            self.assertAlmostEqual(a1.rmin, a2.rmin)
            self.assertAlmostEqual(abs(a1.epsilon), abs(a2.epsilon))
        for a1, a2 in zip(self.sys3.atoms, x3.atoms):
            self.assertAlmostEqual(a1.rmin, a2.rmin)
            self.assertAlmostEqual(abs(a1.epsilon), abs(a2.epsilon))
        # Now try the Amoeba topology
        parm = pmd.load_file(get_fn('nma.parm7'))
        parm.save(self.get_fn('test4.parm7', written=True))
        self.assertIsInstance(pmd.load_file(self.get_fn('test4.parm7', written=True)), pmd.amber.AmoebaParm)

    @unittest.skipUnless(HAS_GROMACS, 'Cannot test without GROMACS')
    def test_save_amber_parm2(self):
        """ Test saving AmberParm with custom exceptions """
        parm = pmd.load_file(os.path.join(get_fn('04.Ala'), 'topol.top'),
                             xyz=os.path.join(get_fn('04.Ala'), 'conf.gro'))
        fn = self.get_fn('test.parm7', written=True)
        parm.save(fn, overwrite=True)
        self.assertIs(type(pmd.load_file(fn)), pmd.amber.AmberParm)
        # Now modify one of the exceptions
        parm.adjust_types.append(copy(parm.adjust_types[0]))
        parm.adjust_types[-1].rmin = 0.1
        parm.adjust_types[-1].epsilon = 0.1
        for adjust in parm.adjusts:
            if adjust.atom1 not in adjust.atom2.dihedral_partners:
                continue
            break
        else:
            assert False, 'No pair types that are 1-4 pairs'
        adjust.type = parm.adjust_types[-1]
        parm.save(fn, overwrite=True)
        # Now it should be a ChamberParm
        self.assertIs(type(pmd.load_file(fn)), pmd.amber.ChamberParm)

    def test_save_psf(self):
        """ Test saving various Structure instances as CHARMM PSF files """
        self.sys1.save(self.get_fn('test.psf', written=True))
        self.sys2.save(self.get_fn('test2.psf', written=True))
        self.sys3.save(self.get_fn('test3.psf', written=True))
        stringio_file = StringIO()
        self.sys3.save(stringio_file, format='psf')
        stringio_file.seek(0)
        x1 = pmd.charmm.CharmmPsfFile(self.get_fn('test.psf', written=True))
        x2 = pmd.charmm.CharmmPsfFile(self.get_fn('test2.psf', written=True))
        x3 = pmd.charmm.CharmmPsfFile(self.get_fn('test3.psf', written=True))
        x4 = pmd.charmm.CharmmPsfFile(stringio_file)
        # PSF files save "improper periodic" torsions under the improper list,
        # only moving them over to the dihedral list once parameters have been
        # assigned and the fact that it's a periodic improper torsion becomes
        # known. Since no parameters are assigned in this section, we need to
        # have a way of making sure that we don't get a false positive based on
        # this difference. Add methods to determine the numbers of proper and
        # improper torsions
        def _propers(struct):
            # Only uniques
            nnormal = 0
            added = set()
            for dih in struct.dihedrals:
                a1, a2, a3, a4 = dih.atom1, dih.atom2, dih.atom3, dih.atom4
                if dih.improper: continue
                if (a1, a2, a3, a4) in added or (a4, a3, a2, a1) in added:
                    continue
                nnormal += 1
                added.add((a1, a2, a3, a4))
            return nnormal
        def _impropers(struct):
            return sum(1 for dih in struct.dihedrals if dih.improper) + len(struct.impropers)
        # Check equivalence of topologies
        self.assertEqual([a.name for a in self.sys1.atoms], [a.name for a in x1.atoms])
        self.assertEqual([a.name for a in self.sys2.atoms], [a.name for a in x2.atoms])
        self.assertEqual([a.name for a in self.sys3.atoms], [a.name for a in x3.atoms])
        self.assertEqual([a.name for a in self.sys3.atoms], [a.name for a in x4.atoms])
        self.assertEqual(len(self.sys1.bonds), len(x1.bonds))
        self.assertEqual(len(self.sys2.bonds), len(x2.bonds))
        self.assertEqual(len(self.sys3.bonds), len(x3.bonds))
        self.assertEqual(len(self.sys1.angles), len(x1.angles))
        self.assertEqual(len(self.sys2.angles), len(x2.angles))
        self.assertEqual(len(self.sys3.angles), len(x3.angles))
        self.assertEqual(_propers(self.sys1), len(x1.dihedrals))
        self.assertEqual(_propers(self.sys2), len(x2.dihedrals))
        self.assertEqual(_propers(self.sys3), len(x3.dihedrals))
        self.assertEqual(_impropers(self.sys1), len(x1.impropers))
        self.assertEqual(_impropers(self.sys2), len(x2.impropers))
        self.assertEqual(_impropers(self.sys3), len(x3.impropers))
        self.assertEqual(len(self.sys1.cmaps), len(x1.cmaps))
        self.assertEqual(len(self.sys2.cmaps), len(x2.cmaps))
        self.assertEqual(len(self.sys3.cmaps), len(x3.cmaps))

    def test_save_charmm_crd(self):
        """ Test saving various Structure instances as CHARMM coord files """
        self.sys1.save(self.get_fn('test.crd', written=True))
        self.sys2.save(self.get_fn('test2.crd', written=True))
        self.sys3.save(self.get_fn('test3.crd', written=True))
        x1 = pmd.charmm.CharmmCrdFile(self.get_fn('test.crd', written=True))
        x2 = pmd.charmm.CharmmCrdFile(self.get_fn('test2.crd', written=True))
        x3 = pmd.charmm.CharmmCrdFile(self.get_fn('test3.crd', written=True))

        np.testing.assert_allclose(self.sys1.coordinates,
                x1.coordinates.reshape(self.sys1.coordinates.shape))
        np.testing.assert_allclose(self.sys2.coordinates,
                x2.coordinates.reshape(self.sys2.coordinates.shape))
        np.testing.assert_allclose(self.sys3.coordinates,
                x3.coordinates.reshape(self.sys3.coordinates.shape))

    def test_save_gromacs(self):
        """ Test saving various Structure instances as GROMACS top files """
        self.sys1.save(self.get_fn('test.top', written=True))
        self.sys2.save(self.get_fn('test2.top', written=True))
        self.sys3.save(self.get_fn('test3.top', written=True))
        stringio_file = StringIO()
        self.sys3.save(stringio_file, format='gromacs')
        stringio_file.seek(0)
        x1 = pmd.gromacs.GromacsTopologyFile(self.get_fn('test.top', written=True))
        x2 = pmd.gromacs.GromacsTopologyFile(self.get_fn('test2.top', written=True))
        x3 = pmd.gromacs.GromacsTopologyFile(self.get_fn('test3.top', written=True))
        x4 = pmd.gromacs.GromacsTopologyFile(stringio_file)
        # Check equivalence of topologies
        self.assertEqual([a.name for a in self.sys1.atoms], [a.name for a in x1.atoms])
        self.assertEqual([a.name for a in self.sys2.atoms], [a.name for a in x2.atoms])
        self.assertEqual([a.name for a in self.sys3.atoms], [a.name for a in x3.atoms])
        self.assertEqual([a.name for a in self.sys3.atoms], [a.name for a in x4.atoms])
        self.assertEqual(len(self.sys1.bonds), len(x1.bonds))
        self.assertEqual(len(self.sys2.bonds), len(x2.bonds))
        self.assertEqual(len(self.sys3.bonds), len(x3.bonds))
        self.assertEqual(len(self.sys1.angles), len(x1.angles))
        self.assertEqual(len(self.sys2.angles), len(x2.angles))
        self.assertEqual(len(self.sys3.angles), len(x3.angles))
        self.assertEqual(len(self.sys1.dihedrals), len(x1.dihedrals))
        self.assertEqual(len(self.sys2.dihedrals), len(x2.dihedrals))
        self.assertEqual(len(self.sys3.dihedrals), len(x3.dihedrals))
        self.assertEqual(len(self.sys1.impropers), len(x1.impropers))
        self.assertEqual(len(self.sys2.impropers), len(x2.impropers))
        self.assertEqual(len(self.sys3.impropers), len(x3.impropers))
        self.assertEqual(len(self.sys1.cmaps), len(x1.cmaps))
        self.assertEqual(len(self.sys2.cmaps), len(x2.cmaps))
        self.assertEqual(len(self.sys3.cmaps), len(x3.cmaps))

    def test_save_psf2(self):
        """ Test saving PSF file for unparametrized system """
        url = 'http://ambermd.org/tutorials/advanced/tutorial1/files/polyAT.pdb'
        pmd.load_file(url).save(self.get_fn('test.psf', written=True))

    def test_save_gro(self):
        """ Test saving various Structure instances as a PDB """
        self.sys1.save(self.get_fn('test.gro', written=True))
        self.sys2.save(self.get_fn('test2.gro', written=True))
        self.sys3.save(self.get_fn('test3.gro', written=True))
        stringio_file = StringIO()
        self.sys3.save(stringio_file, format='gro')
        stringio_file.seek(0)
        x1 = pmd.gromacs.GromacsGroFile.parse(self.get_fn('test.gro', written=True))
        x2 = pmd.gromacs.GromacsGroFile.parse(self.get_fn('test2.gro', written=True))
        x3 = pmd.gromacs.GromacsGroFile.parse(self.get_fn('test3.gro', written=True))
        x4 = pmd.gromacs.GromacsGroFile.parse(stringio_file)
        self.assertEqual([a.name for a in self.sys1.atoms], [a.name for a in x1.atoms])
        self.assertEqual([a.name for a in self.sys2.atoms], [a.name for a in x2.atoms])
        self.assertEqual([a.name for a in self.sys3.atoms], [a.name for a in x3.atoms])
        self.assertEqual([a.name for a in self.sys3.atoms], [a.name for a in x4.atoms])

    def test_save_rst7(self):
        """ Test saving various Structure instances as Amber ASCII restarts """
        f1 = self.get_fn('test.rst7', written=True)
        f2 = self.get_fn('test1.restrt', written=True)
        f3 = self.get_fn('test2.inpcrd', written=True)
        f4 = self.get_fn('test3.amberrst', written=True)
        stringio_file = StringIO()
        self.sys1.save(f1)
        self.sys2.save(f2)
        self.sys3.save(f3)
        self.sys1.save(f4, format='rst7')
        self.sys1.save(stringio_file, format='rst7')
        stringio_file.seek(0)

        self.assertTrue(pmd.amber.AmberAsciiRestart.id_format(f1))
        self.assertTrue(pmd.amber.AmberAsciiRestart.id_format(f2))
        self.assertTrue(pmd.amber.AmberAsciiRestart.id_format(f3))
        self.assertTrue(pmd.amber.AmberAsciiRestart.id_format(f4))
        self.assertTrue(pmd.amber.AmberAsciiRestart.id_format(stringio_file))

        np.testing.assert_allclose(self.sys1.coordinates,
                                   pmd.load_file(f1).coordinates[0],
                                   atol=1e-6)
        np.testing.assert_allclose(self.sys2.coordinates,
                                   pmd.load_file(f2).coordinates[0],
                                   atol=1e-6)
        np.testing.assert_allclose(self.sys3.coordinates,
                                   pmd.load_file(f3).coordinates[0],
                                   atol=1e-6)
        np.testing.assert_allclose(self.sys1.coordinates,
                                   pmd.load_file(f4).coordinates[0],
                                   atol=1e-6)

    @unittest.skipIf(PYPY, 'NetCDF tests cannot run on pypy yet')
    def test_save_ncrst7(self):
        """ Test saving various Structure instances as Amber NetCDF restarts """
        f1 = self.get_fn('test.ncrst', written=True)
        f2 = self.get_fn('test1.ncrst', written=True)
        f3 = self.get_fn('test2.ncrestart', written=True)
        self.sys1.save(f1)
        self.sys2.save(f2)
        self.sys3.save(f3, format='ncrst')

        self.assertTrue(pmd.amber.NetCDFRestart.id_format(f1))
        self.assertTrue(pmd.amber.NetCDFRestart.id_format(f2))
        self.assertTrue(pmd.amber.NetCDFRestart.id_format(f3))

        np.testing.assert_allclose(self.sys1.coordinates,
                                   pmd.load_file(f1).coordinates[0])
        np.testing.assert_allclose(self.sys2.coordinates,
                                   pmd.load_file(f2).coordinates[0])
        np.testing.assert_allclose(self.sys3.coordinates,
                                   pmd.load_file(f3).coordinates[0])

    def test_save_pqr(self):
        """ Test saving various Structure instances as PQR files """
        f1 = self.get_fn('test', written=True)
        f2 = self.get_fn('test.pqr', written=True)
        f3 = self.get_fn('test2.pqr', written=True)
        self.sys1.save(f1, format='pqr')
        self.sys2.save(f2)
        self.sys3.save(f3)

        self.assertTrue(pmd.formats.PQRFile.id_format(f1))
        self.assertTrue(pmd.formats.PQRFile.id_format(f2))
        self.assertTrue(pmd.formats.PQRFile.id_format(f3))

        x1 = pmd.formats.PQRFile.parse(f1)
        x2 = pmd.formats.PQRFile.parse(f2)
        x3 = pmd.formats.PQRFile.parse(f3)

        self.assertEqual([a.name for a in self.sys1.atoms], [a.name for a in x1.atoms])
        self.assertEqual([a.name for a in self.sys2.atoms], [a.name for a in x2.atoms])
        self.assertEqual([a.name for a in self.sys3.atoms], [a.name for a in x3.atoms])

        np.testing.assert_allclose(np.array([a.charge for a in self.sys1.atoms]),
                                   np.array([a.charge for a in x1.atoms]))

        np.testing.assert_allclose(np.array([a.charge for a in self.sys2.atoms]),
                                   np.array([a.charge for a in x2.atoms]))

        np.testing.assert_allclose(np.array([a.charge for a in self.sys3.atoms]),
                                   np.array([a.charge for a in x3.atoms]))

    def test_error_handling(self):
        """ Tests handling of bad input """
        self.assertRaises(ValueError, lambda:
                self.sys1.save('somefile', format='NOFMT'))
        self.assertRaises(ValueError, lambda: self.sys1.save('somefile.nofmt'))

    def test_overwrite(self):
        """ Test overwrite option of Structure.save """
        open(self.get_fn('test.pdb', written=True), 'w').close()
        with self.assertRaises(IOError):
            self.sys1.save(self.get_fn('test.pdb', written=True))
        # Should not raise
        self.sys1.save(self.get_fn('test.pdb', written=True), overwrite=True)
        pdb = pmd.load_file(self.get_fn('test.pdb', written=True))
        self.assertEqual(len(pdb.atoms), len(self.sys1.atoms))



"""
A little utility for performing C-like preprocessing using some standard CPP
directives like #if, #ifdef, and #define.

Written by Jason Swails
"""
import re
import warnings
from functools import wraps
from collections import OrderedDict
from os import path
from ..exceptions import PreProcessorError, PreProcessorWarning
from ..utils.io import genopen

ppre = re.compile(r'#\s*(ifdef|ifndef|if|else|elif|endif|define|undef|include)\s*(.+)?')
ppcomments = re.compile(r'(?://.+|/\*(?:.*)\*/)')
includere = re.compile(r'[<"](.+)[>"]')
novarcharre = re.compile(r'\W')

def _strip_pp_comments(func):
    """
    Decorator to apply to functions that will strip out C-style comments before
    calling the preprocessor function on the resulting arguments to the
    preprocessor directive
    """
    @wraps(func)
    def wrapper(self, args):
        args = ppcomments.sub('', args)
        return func(self, args)
    return wrapper

def _find_all_instances_in_string(string, substr):
    """ Find indices of all instances of substr in string """
    indices = []
    idx = string.find(substr, 0)
    while idx > -1:
        indices.append(idx)
        idx = string.find(substr, idx+1)
    return indices

def _replace_defines(line, defines):
    """ Replaces defined tokens in a given line """
    if not defines: return line
    for define in reversed(defines):
        value = defines[define]
        indices = _find_all_instances_in_string(line, define)
        if not indices: continue
        # Check to see if it's inside of quotes
        inside = ''
        idx = 0
        n_to_skip = 0
        new_line = []
        for i, char in enumerate(line):
            if n_to_skip:
                n_to_skip -= 1
                continue
            if char in ('\'"'):
                if not inside:
                    inside = char
                else:
                    if inside == char:
                        inside = ''
            if idx < len(indices) and i == indices[idx]:
                if inside:
                    new_line.append(char)
                    idx += 1
                    continue
                if i == 0 or novarcharre.match(line[i-1]):
                    endidx = indices[idx] + len(define)
                    if endidx >= len(line) or novarcharre.match(line[endidx]):
                        new_line.extend(list(value))
                        n_to_skip = len(define) - 1
                        idx += 1
                        continue
                idx += 1
            new_line.append(char)
        line = ''.join(new_line)
                        
    return line

# To track where in the "if-elif-else" block each conditional is
_IN_IF = 'in if'
_IN_ELIF = 'in elif'
_IN_ELSE = 'in else'

class CPreProcessor(object):
    """
    Steps through a file line-by-line, yielding only the preprocessed contents

    Parameters
    ----------
    fname : str or file-like
        The file name of the file to open or the open file object
    defines : dict{str : value}, optional
        The dict of defines to apply to this preprocessed file if any;
        equivalent to the list of -DXYZ=ABC in the standard preprocessor
    includes : list of str, optional
        List of directory names to search for included files, if any
    notfound_fatal : bool, optional
        If True, include files not found are fatal. If False, they will simply
        be skipped (with a warning emitted). Default True

    Notes
    -----
    If ``fname`` is a file name, the directory containing that file name is the
    first directory searched for include files. If ``fname`` is a file-like
    object, then the current directory is the first searched.
    """

    def __init__(self, fname, defines=None, includes=None, notfound_fatal=True):
        if isinstance(fname, str):
            self._fileobj = genopen(fname, 'r')
            self._ownhandle = True
            curpath = path.abspath(path.split(fname)[0])
            self.filename = fname
        else:
            self._fileobj = fname
            self._ownhandle = False
            curpath = path.abspath(path.curdir)
            self.filename = None
        if includes is None:
            self._includes = [curpath]
        else:
            self._includes = [curpath] + list(includes)
        if defines is None:
            self.defines = OrderedDict()
        else:
            # Convert every define to a string
            self.defines = OrderedDict()
            for define, value in defines.items():
                self.defines[define] = str(value)
        self._notfound_fatal = notfound_fatal

        # Now to keep track of other basic logic stuff
        self.included_files = []
        self._ifstack = []
        self._elsestack = []
        self._satisfiedstack = []
        self._num_ignoring_if = 0
        self._includefile = None

    def __del__(self):
        self.close()

    def close(self):
        if hasattr(self, '_ownhandle') and self._ownhandle:
            self._fileobj.close()

    def readline(self):
        try:
            return next(iter(self))
        except StopIteration:
            return ''

    def readlines(self):
        return [line for line in self]

    def read(self):
        return ''.join(self.readlines())

    def tell(self):
        return self._fileobj.tell()

    def seek(self, value):
        raise NotImplementedError('Cannot seek through a preprocessed file')

    def __iter__(self):
        """ Step through the preprocessed file line-by-line """
        for line in self._fileobj:
            rematch = ppre.match(line)
            if rematch:
                cmd, args = rematch.groups()
                args = args or ''
                self._ppcmdmap[cmd](self, args)
                # If we defined an include file, step through it
                if self._includefile is not None:
                    for line in self._includefile:
                        yield line
                    self._includefile.close()
                    # We have to pass our defines back to our caller
                    self.defines = self._includefile.defines
                    self._includefile = None
                continue

            if self._satisfiedstack and not self._satisfiedstack[-1]:
                # We are inside an unsatisfied conditional
                continue

            yield _replace_defines(line, self.defines)
        # Make sure we don't have any dangling ifs
        if self._ifstack:
            raise PreProcessorError('EOF: Unterminated #if(def)')

    @_strip_pp_comments
    def _pp_if(self, args):
        if self._satisfiedstack and not self._satisfiedstack[-1]:
            self._num_ignoring_if += 1
            return
        args = args.strip()
        if args.strip() in ('0', '1'):
            satisfied = bool(int(args))
            self._ifstack.append(str(bool(satisfied)))
            self._satisfiedstack.append(bool(satisfied))
            self._elsestack.append(_IN_IF)
            return
        raise NotImplementedError('Only "#if 0|1" is currently supported')

    @_strip_pp_comments
    def _pp_ifdef(self, args):
        if self._satisfiedstack and not self._satisfiedstack[-1]:
            self._num_ignoring_if += 1
            return
        words = args.split()
        if len(words) == 0:
            raise PreProcessorError(f'Bad #ifdef syntax: "#ifdef {args}"')
        elif len(words) > 1:
            warnings.warn("Ignored tokens in #ifdef: {', '.join(words[1:])}", PreProcessorWarning)
        self._ifstack.append(f'{words[0]} in self.defines')
        self._elsestack.append(_IN_IF)
        self._satisfiedstack.append(words[0] in self.defines)

    @_strip_pp_comments
    def _pp_ifndef(self, args):
        if self._satisfiedstack and not self._satisfiedstack[-1]:
            self._num_ignoring_if += 1
            return
        words = args.split()
        if len(words) == 0:
            raise PreProcessorError(f'Bad #ifndef syntax: "#ifndef {args}"')
        elif len(words) > 1:
            warnings.warn(f"Ignored tokens in #ifndef: {', '.join(words[1:])}", PreProcessorWarning)
        self._ifstack.append(f'{words[0]} not in self.defines')
        self._elsestack.append(_IN_IF)
        self._satisfiedstack.append(words[0] not in self.defines)

    @_strip_pp_comments
    def _pp_elif(self, args):
        raise NotImplementedError('#elif conditionals are not yet implemented.')

    @_strip_pp_comments
    def _pp_else(self, args):
        if self._num_ignoring_if > 0:
            return
        if not self._ifstack:
            raise PreProcessorError('#else missing #if(def)')
        words = args.split()
        if len(words) > 0:
            warnings.warn(f"Ignored tokens in #else: {', '.join(words[1:])}", PreProcessorWarning)
        if self._elsestack[-1] == _IN_ELSE:
            raise PreProcessorError('#else following #else')
        self._elsestack[-1] = _IN_ELSE
        self._satisfiedstack[-1] = not self._satisfiedstack[-1]

    @_strip_pp_comments
    def _pp_endif(self, args):
        if self._num_ignoring_if > 0:
            self._num_ignoring_if -= 1
            return
        if args.strip():
            warnings.warn(f'Ignored tokens in #endif: {args.strip()}', PreProcessorWarning)
        if not self._ifstack:
            raise PreProcessorError('#endif missing #if(def)')
        self._ifstack.pop()
        self._satisfiedstack.pop()
        self._elsestack.pop()

    @_strip_pp_comments
    def _pp_include(self, args):
        if self._satisfiedstack and not self._satisfiedstack[-1]:
            return
        # Locate the include file
        rematch = includere.match(args)
        if not rematch:
            raise PreProcessorError('Bad #include syntax')
        includefile = rematch.groups()[0]
        self.included_files.append(includefile)
        for folder in self._includes:
            testfile = path.join(folder, includefile)
            if path.isfile(testfile):
                break
        else:
            if self._notfound_fatal:
                raise PreProcessorError(f'Could not find {includefile}')
            warnings.warn(f'Could not find {includefile}; skipping', PreProcessorWarning)
            return
        self._includefile = CPreProcessor(testfile,
                                          defines=self.defines,
                                          includes=self._includes,
                                          notfound_fatal=self._notfound_fatal)

    @_strip_pp_comments
    def _pp_define(self, args):
        if self._satisfiedstack and not self._satisfiedstack[-1]:
            return
        # Define a new variable
        words = args.split()
        if len(words) == 0:
            raise PreProcessorError('Nothing defined in #define')
        # Warn about a double-define
        if words[0] in self.defines:
            warnings.warn(f'{words[0]} already defined; overwriting', PreProcessorWarning)
        if len(words) == 1:
            self.defines[words[0]] = '1'
        elif len(words) >= 2:
            self.defines[words[0]] = args[len(words[0]):].strip()

    @_strip_pp_comments
    def _pp_undef(self, args):
        if self._satisfiedstack and not self._satisfiedstack[-1]:
            return
        # Define a new variable
        words = args.split()
        if len(words) == 1:
            try:
                del self.defines[words[0]]
            except KeyError:
                # Undefining an undefined variable is a no-op
                pass
        elif len(words) > 1:
            warnings.warn(f"Ignored tokens in #undef: {', '.join(words[1:])}", PreProcessorWarning)
        elif len(words) == 0:
            raise PreProcessorError('Nothing defined in #undef')

    # Context manager protocol
    def __exit__(self, type, value, traceback):
        self.close()

    def __enter__(self):
        return self

    _ppcmdmap = {'if' : _pp_if, 'elif' : _pp_elif, 'ifdef' : _pp_ifdef,
                 'else' : _pp_else, 'define' : _pp_define, 'undef' : _pp_undef,
                 'include' : _pp_include, 'endif' : _pp_endif,
                 'ifndef' : _pp_ifndef}

if __name__ == '__main__':
    # Act as a stand-alone preprocessor
    import argparse
    import sys
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input-file', dest='input', metavar='FILE',
                required=True, help='''Input file to pre-process. Either a file
                name or, if '-' is given, from standard input.''')
    parser.add_argument('-o', '--output-file', dest='output', metavar='FILE',
                default=None, help='''Output file with preprocessed results.
                Default is standard output''')
    parser.add_argument('-D', dest='defines', metavar='VAR[=VAL]',
                action='append', help='''List of predefined variables to pass to
                the preprocessor. Default VAL is 1 when missing.''', default=[])
    parser.add_argument('-I', dest='includes', metavar='DIRECTORY',
                action='append', help='''List of include directories to search
                for included files''', default=[])

    opt = parser.parse_args()

    defines = OrderedDict()
    for define in opt.defines:
        if '=' in define:
            define, val = define.split('=')
        else:
            val = '1'
        defines[define] = val

    if opt.input == '-':
        f = sys.stdin
    else:
        f = opt.input
    pp = CPreProcessor(f, defines=defines, includes=opt.includes)
    if opt.output is None:
        output = sys.stdout
        own_handle = False
    else:
        output = genopen(opt.output, 'w')
        own_handle = True

    for line in pp:
        output.write(line)
    if own_handle:
        output.close()



"""
Tests the functionality in the parmed.unit package.
"""
import utils
import numpy as np
from parmed import unit as u
import copy
import math
import unittest
import utils

class TestUnits(utils.QuantityTestCase):

    def testBaseUnit(self):
        """ Tests the creation of a base unit furlong """
        furlong = u.BaseUnit(u.length_dimension, "furlong", "fur")
        furlong.define_conversion_factor_to(u.meter_base_unit, 201.168)
        self.assertEqual(furlong.conversion_factor_to(u.angstrom_base_unit), 201.168e10)

    def testIsUnit(self):
        """ Tests the unit.is_unit introspective function """
        self.assertTrue(u.is_unit(u.meters))
        self.assertFalse(u.is_unit(None))
        self.assertFalse(u.is_unit(10))
        # Distinguish between "units" and "quantities"
        self.assertFalse(u.is_unit(1*u.meters))

    def testCalorieConversion(self):
        """ Tests unit conversion with calories to joules """
        c = 1.0 * u.calories
        j = 1.0 * u.joules
        j2c = j.in_units_of(u.calories)
        self.assertNotEqual(c, 1.0) # Units do not equal scalars!
        self.assertIs(c.unit, u.calories)
        self.assertEqual(c._value, 1)
        self.assertEqual(u.calorie.conversion_factor_to(u.joule), 4.184)
        self.assertEqual(u.joule.conversion_factor_to(u.calorie), 1/4.184)
        self.assertIs(j2c.unit, u.calories)
        self.assertEqual(j2c._value, 1/4.184)
        self.assertEqual(c / u.calories, 1.0) # Dimensionless now
        self.assertEqual(j / u.calories, j2c._value)
        c2 = c**2
        cc = c * c
        self.assertEqual(cc, c2)
        self.assertTrue(c2.unit, u.calories**2)
        self.assertEqual(c2.in_units_of(u.joules**2), (4.184*u.joules)**2)

    def testScaledUnit(self):
        """ Tests ScaledUnit class with kilokelvins """
        kK = u.ScaledUnit(1000.0, u.kelvin, "kilokelvin", "kK")
        self.assertIs(kK.master, u.kelvin)
        self.assertEqual(kK.factor, 1000)
        self.assertEqual(str(kK), 'kilokelvin')
        self.assertEqual(kK.name, 'kilokelvin')
        self.assertEqual(kK.symbol, 'kK')

    def testComparisonOperators(self):
        """ Tests unit comparison operators """
        self.assertGreater(u.meters, u.centimeters)
        self.assertLess(u.angstroms, u.centimeters)

    def testUnitDivision(self):
        """ Tests the division of units and is_dimensionless """
        mps = u.meter / u.second
        mpm = u.meter / u.meter
        mpc = u.meter / u.centimeter
        self.assertTrue(u.is_unit(mps))
        self.assertTrue(u.is_unit(mpm))
        self.assertTrue(u.is_unit(mpc))
        self.assertFalse(mps.is_dimensionless())
        self.assertTrue(mpm.is_dimensionless())
        self.assertTrue(mpc.is_dimensionless())

    def testCompositeUnits(self):
        """ Tests the creation of a composite unit """
        mps = u.Unit({u.meter_base_unit : 1.0, u.second_base_unit : -1.0})
        self.assertTrue(u.is_unit(mps))
        self.assertEqual(str(mps), 'meter/second')

    def testUnitSystem(self):
        """ Tests the creation of a UnitSystem and its behavior """
        us = u.UnitSystem([u.ScaledUnit(1.0, u.coulomb/u.second, 'ampere', 'A'), u.second_base_unit])
        self.assertEqual(us.express_unit(u.second), u.second)
        self.assertEqual(us.express_unit(u.hour), u.second)
        self.assertNotEqual(u.hour, u.second)
        self.assertEqual(us.express_unit(u.coulomb/u.second), u.ampere)
        self.assertEqual(us.express_unit(u.meter/u.second), u.meter/u.second)
        self.assertEqual(us.express_unit(u.kilometer/u.hour), u.kilometer/u.second)
        x = 100 * u.millimeters
        self.assertEqual(x.value_in_unit_system(u.si_unit_system), 0.1)
        self.assertEqual(x.value_in_unit_system(u.cgs_unit_system), 10.0)
        self.assertAlmostEqual(x.value_in_unit_system(u.md_unit_system), 1e8)
        y = 20 * u.millimeters / u.millisecond**2
        self.assertAlmostEqual(y.value_in_unit_system(u.si_unit_system), 2e4)
        self.assertAlmostEqual(y.value_in_unit_system(u.cgs_unit_system), 2e6)
        self.assertAlmostEqual(y.value_in_unit_system(u.md_unit_system), 2e-11, places=18)
        kcal = 1 * u.md_kilocalorie / u.mole
        self.assertEqual(kcal.value_in_unit_system(u.md_unit_system), 4.184)

    def testUnitSqrt(self):
        """ Tests taking the square root of units """
        self.assertEqual((u.meter * u.meter).sqrt(), u.meter)
        self.assertEqual((u.meter**4).sqrt(), u.meter**2)

    def testQuantitySqrt(self):
        """ Tests taking the square root of quantities """
        self.assertEqual((9.0*u.meters**2).sqrt(), 3.0*u.meters)
        self.assertEqual((9.0*u.meters**2/u.second**2).sqrt(), 3.0*u.meters/u.second)

    def testUnitBadSqrt(self):
        """ Tests that illegal sqrt calls on incompatible units fails """
        mps2 = u.meters/u.second**2
        self.assertRaises(ArithmeticError, lambda: u.meter.sqrt())
        self.assertRaises(ArithmeticError, lambda: (u.meters**3).sqrt())
        self.assertRaises(ArithmeticError, lambda: mps2.sqrt())

    def testQuantityBadSqrt(self):
        " Tests that taking sqrt of Quantities with incompatible units fails "
        self.assertRaises(ArithmeticError, lambda: (9.0*u.meters).sqrt())
        self.assertRaises(ArithmeticError, lambda: (9.0*u.meters**3).sqrt())

    def testBaseScaleMix(self):
        """ Test mixing of BaseUnit and ScaledUnit instances """
        kgj = u.kilogram * u.joule
        self.assertTrue(u.is_unit(kgj))
        self.assertEqual(str(kgj.sqrt()), 'kilogram*meter/second')

    def testGetUnitAttributes(self):
        """ Tests the unit attribute `getters' """
        self.assertEqual(u.newton.get_name(), 'newton')
        self.assertEqual(u.newtons.get_name(), 'newton')
        self.assertEqual(u.newton.get_symbol(), 'N')
        self.assertEqual(u.ampere.get_symbol(), 'A')
        self.assertEqual(u.meter.get_name(), 'meter')
        self.assertEqual(u.meter.get_symbol(), 'm')

    def testPresetUnitSystems(self):
        """ Tests some of the pre-set UnitSystem's """
        self.assertEqual(u.angstrom.in_unit_system(u.si_unit_system), u.meter)
        self.assertEqual(u.angstrom.in_unit_system(u.cgs_unit_system), u.centimeter)
        self.assertEqual(u.angstrom.in_unit_system(u.md_unit_system), u.nanometer)
        mps = u.meter / u.second**2
        self.assertEqual(str(mps), 'meter/(second**2)')
        self.assertEqual(mps.in_unit_system(u.si_unit_system), mps)
        self.assertEqual(mps.in_unit_system(u.cgs_unit_system), u.centimeter/u.second**2)
        self.assertEqual(mps.in_unit_system(u.md_unit_system), u.nanometer/u.picosecond**2)

    def testIsCompatible(self):
        """ Tests the is_compatible attribute of units """
        self.assertTrue(u.meter.is_compatible(u.centimeter))
        self.assertTrue(u.centimeter.is_compatible(u.meter))
        self.assertTrue(u.meter.is_compatible(u.meter))
        self.assertTrue(u.joule.is_compatible(u.calorie))
        self.assertFalse(u.meter.is_compatible(u.kelvin))
        self.assertFalse(u.meter.is_compatible(u.meter/u.second))
        self.assertFalse(u.meter.is_compatible(u.joule))

    def testConversionFactorTo(self):
        """ Tests the "conversion_factor_to" attribute of Units """
        self.assertEqual(u.meter.conversion_factor_to(u.centimeter), 100)
        self.assertEqual(u.kilocalorie.conversion_factor_to(u.joule), 4184)
        self.assertEqual(u.calorie.conversion_factor_to(u.kilojoule), 4.184e-3)
        self.assertEqual((u.kilocalorie/u.mole/u.angstrom).conversion_factor_to(
                                u.kilojoule/u.mole/u.nanometer), 41.84)

    def testUnitString(self):
        """ Test the Unit->str casting functionality """
        # Always alphabetical order
        self.assertEqual(str(u.meter * u.second * u.second * u.kilogram), 'kilogram*meter*second**2')
        self.assertEqual(str(u.meter / u.second / u.second / u.kilogram), 'meter/(kilogram*second**2)')
        self.assertEqual(str(u.meter**3), 'meter**3')

    def testToBaseUnit(self):
        """ Tests the "get_conversion_factor_to_base_units" method """
        self.assertEqual(u.meter.get_conversion_factor_to_base_units(), 1)
        self.assertEqual(u.calorie.get_conversion_factor_to_base_units(), 4.184)
        kcpma = u.md_kilocalorie/u.mole/u.angstrom
        self.assertEqual(kcpma.get_conversion_factor_to_base_units(), 4.184)

    def testScalarQuantityMultiplyDivide(self):
        """ Tests creating a scalar Quantity object by * or / by a Unit """
        self.assertTrue(u.is_quantity(5 * u.centimeters))
        self.assertTrue(u.is_quantity(1 / u.centimeters))
        self.assertTrue(u.is_quantity(10 * u.centimeters))
        self.assertTrue(u.is_quantity(9.81 * u.meters / u.second**2))

    def testScalarQuantityConstructor(self):
        """ Tests creating a Quantity using the Quantity constructor """
        self.assertTrue(u.is_quantity(u.Quantity(5, u.centimeters)))
        self.assertTrue(u.is_quantity(u.Quantity(5, u.centimeters**-1)))
        x = u.Quantity(value=5.0, unit=100.0*u.meters)
        self.assertTrue(u.is_quantity(x))
        self.assertEqual(x, 500*u.meters)

    def testValueInUnit(self):
        """ Tests the value_in_unit functionality for Quantity """
        i = 5 * u.centimeters
        self.assertEqual(i.value_in_unit(u.millimeters), 50)
        self.assertEqual(i / u.millimeters, 50)

    def testCollectionQuantities(self):
        """ Tests the use of collections as Quantity values """
        s = [1, 2, 3] * u.centimeters
        self.assertEqual(str(s), '[1, 2, 3] cm')
        self.assertTrue(u.is_quantity(s))
        s2 = s / u.millimeters
        self.assertEqual(s2, [10.0, 20.0, 30.0])
        self.assertEqual(s2, s.value_in_unit(u.millimeters))
        # Test 2-D list
        s = [[1, 2, 3], [4, 5, 6]]
        s *= u.centimeters
        self.assertTrue(u.is_quantity(s))
        s2 = s / u.millimeters
        self.assertEqual(s2, [[10.0, 20.0, 30.0], [40.0, 50.0, 60.0]])
        self.assertEqual(s.value_in_unit(u.millimeters), s2)
        # Test tuples
        s = (1, 2, 3) * u.centimeters
        self.assertTrue(u.is_quantity(s))
        self.assertEqual(str(s), '(1, 2, 3) cm')
        s2 = s / u.millimeters
        self.assertEqual(s2, (10, 20, 30))
        self.assertIsInstance(s2, tuple)
        self.assertEqual(s.value_in_unit(u.millimeters), s2)
        self.assertIsInstance(s.value_in_unit(u.millimeters), tuple)
        x = [1, 2, 3] * u.centimeters
        x *= u.meters
        self.assertEqual(x, [100, 200, 300] * u.centimeters**2)

    def testReduceUnit(self):
        """ Tests the reduce_unit functionality """
        x = u.nanometer**2 / u.angstrom**2
        self.assertEqual(str(x), 'nanometer**2/(angstrom**2)')
        self.assertTrue(x.is_dimensionless())
        q = u.Quantity(2.0, x)
        self.assertEqual(str(q), '2.0 nm**2/(A**2)')
        self.assertEqual(q.reduce_unit(), 200)

    def testCollectionQuantityOperations(self):
        """ Tests that Quantity collections behave correctly """
        # Tests that __getitem__ returns a unit
        s = [1, 2, 3, 4] * u.angstroms
        self.assertTrue(u.is_quantity(s[0]))
        for i, val in enumerate(s):
            self.assertTrue(u.is_quantity(val))
            self.assertEqual(val, (i+1) * u.angstroms)
        # Tests that __setitem__ fails when an incompatible type is added
        def fail(s): s[0] = 5
        self.assertRaises(AttributeError, lambda: fail(s))
        def fail(s): s[0] = 5 * u.joules
        self.assertRaises(TypeError, lambda: fail(s))
        def fail(s): s[0] /= 10 * u.meters
        self.assertRaises(AttributeError, lambda: fail(s))
        # Tests that __setitem__ converts to the unit of the container
        s[0] = 1 * u.nanometers
        self.assertEqual(s[0]._value, 10)
        # Tests standard unit conversions
        x = [1, 2, 3] * u.centimeters
        self.assertEqual(x / u.millimeters, [10, 20, 30])
        # Test the construction of a container in which each element is a
        # Quantity, passed to the Quantity constructor
        x = u.Quantity([1*u.angstrom, 2*u.nanometer, 3*u.angstrom])
        self.assertEqual(x._value, [1, 20, 3])
        self.assertEqual(x.unit, u.angstrom)
        x = u.Quantity((1, 2, 3))
        self.assertTrue(u.is_quantity(x))
        self.assertTrue(x.unit.is_dimensionless())
        x = u.Quantity(([1*u.angstrom, 2*u.nanometer, 3*u.angstrom],
                        [1*u.angstrom, 4*u.nanometer, 3*u.angstrom]))
        self.assertEqual(x._value, ([1, 20, 3], [1, 40, 3]))
        self.assertEqual(x.unit, u.angstrom)
        self.assertTrue(u.is_quantity(u.Quantity([])))

    def testMutableQuantityOperations(self):
        " Tests that mutable Quantity objects do not get unexpectedly changed "
        # This used to be a bug -- t and s._value were the same object, so
        # changing t would also change s silently
        s = [1, 2, 3, 4] * u.angstroms
        t = s / u.angstroms
        self.assertEqual(t, [1, 2, 3, 4])
        self.assertEqual(s, u.Quantity([1, 2, 3, 4], u.angstroms))
        t[0] = 2
        self.assertEqual(t, [2, 2, 3, 4])
        self.assertEqual(s, u.Quantity([1, 2, 3, 4], u.angstroms))
        t = s.value_in_unit(u.angstroms)
        self.assertEqual(t, [1, 2, 3, 4])
        self.assertEqual(s, u.Quantity([1, 2, 3, 4], u.angstroms))
        t[0] = 2
        self.assertEqual(t, [2, 2, 3, 4])
        self.assertEqual(s, u.Quantity([1, 2, 3, 4], u.angstroms))
        s = [1, 2, 3, 4] * u.nanometers
        self.assertEqual(s, u.Quantity([1, 2, 3, 4], u.nanometers))
        t = s.in_units_of(u.nanometers)
        self.assertEqual(s, t)
        t[0] = 1 * u.meters
        self.assertAlmostEqualQuantities(t, u.Quantity([1e9, 2, 3, 4], u.nanometers))
        self.assertEqual(s, u.Quantity([1, 2, 3, 4], u.nanometers))

    def testQuantityMaths(self):
        """ Tests dimensional analysis & maths on and b/w Quantity objects """
        x = 1.3 * u.meters
        y = 75.2 * u.centimeters
        self.assertEqual((x + y) / u.meters, 2.052)
        self.assertEqual((x - y) / u.meters, 0.548)
        self.assertEqual(x / y, 1.3 / 0.752)
        self.assertEqual(x * y, 1.3*0.752*u.meters**2)
        d1 = 2.0*u.meters
        d2 = 2.0*u.nanometers
        self.assertEqual(d1 + d2, (2+2e-9)*u.meters)
        self.assertAlmostEqual((d2+d1-(2e9+2)*u.nanometers)._value, 0, places=6)
        self.assertEqual(d1 + d1, 4.0*u.meters)
        self.assertEqual(d1 - d1, 0.0*u.meters)
        self.assertEqual(d1 / d1, 1.0)
        self.assertEqual(d1 * u.meters, 2.0*u.meters**2)
        self.assertEqual(u.kilograms*(d1/u.seconds)*(d1/u.seconds), 4*u.kilograms*u.meters**2/u.seconds**2)
        self.assertEqual(u.kilograms*(d1/u.seconds)**2, 4*u.kilograms*u.meters**2/u.seconds**2)
        self.assertEqual(d1**3, 8.0*u.meters**3)
        x = d1**(3/2)
        self.assertAlmostEqual(x._value, math.sqrt(2)**3)
        self.assertEqual(x.unit, u.meters**(3/2))
        self.assertAlmostEqual((d1**0.5)._value, math.sqrt(2))
        self.assertEqual((d1**0.5).unit, u.meters**0.5)
        comp = (3.0 + 4.0j) * u.meters
        self.assertTrue(u.is_quantity(comp))
        self.assertEqual(comp.unit, u.meters)
        self.assertEqual(str(comp), '(3+4j) m')
        self.assertEqual(comp + comp, (6.0 + 8.0j)*u.meters)
        self.assertEqual(comp - comp, 0*u.meters)
        self.assertEqual(comp * comp, (3.0 + 4.0j)**2 * u.meters**2)
        self.assertAlmostEqual(abs(comp / comp), 1)
        self.assertAlmostEqual(1.5*u.nanometers / u.meters, 1.5e-9, places=15)
        self.assertEqual((2.3*u.meters)**2, 2.3**2*u.meters**2)
        x = 4.3 * u.meters
        self.assertEqual(x / u.centimeters, 430)
        self.assertEqual(str(x / u.seconds), '4.3 m/s')
        self.assertEqual(str(8.4 / (4.2*u.centimeters)), '2.0 /cm')
        x = 1.2 * u.meters
        self.assertEqual(x * 5, u.Quantity(6.0, u.meters))

    def testQuantityComplicatedMaths(self):
        """ Tests a complicated set of mathematical operations on a Quantity """
        s1 = 2.0
        x1 = 2
        x2 = 4.0 / 3.0
        u1 = u.kilogram * u.meter / u.second**2
        u2 = u1 * u.meter
        q1 = 1.0 * u1
        q2 = 2.0 * u2
        self.assertEqual(s1, 2.0)
        self.assertEqual(x1, 2)
        self.assertAlmostEqual(x2, 1.33333333333333)
        self.assertEqual(str(u1), 'kilogram*meter/(second**2)')
        self.assertEqual(str(u2), 'kilogram*meter**2/(second**2)')
        self.assertEqual(str(q1), '1.0 kg m/(s**2)')
        self.assertEqual(str(q2), '2.0 kg m**2/(s**2)')
        self.assertEqual(str(u1*s1), '2.0 kg m/(s**2)')
        self.assertEqual(str(s1*u1), '2.0 kg m/(s**2)')
        self.assertEqual(str(u1/s1), '0.5 kg m/(s**2)')
        self.assertEqual(str(s1/u1), '2.0 s**2/(kg m)')
        self.assertEqual(str(u1*u1), 'kilogram**2*meter**2/(second**4)')
        self.assertEqual(u1/u1, u.dimensionless)
        self.assertEqual(str(u1/u1), 'dimensionless')
        self.assertEqual(str(u1*u2), 'kilogram**2*meter**3/(second**4)')
        self.assertEqual(u1/u2, u.meters**-1)
        self.assertEqual(str(u1/u2), '/meter')
        self.assertEqual(u1**x1, u.kilogram**2*u.meter**2/(u.second**4))
        self.assertEqual(u1**(1/x1), u.kilogram**0.5*u.meter**0.5/u.second)
        self.assertEqual(u1**x2, u.kilogram**(1+1/3)*u.meter**(1+1/3)/u.second**(2+2/3))
        q = 1.0 * u.md_kilocalorie/u.mole/u.angstrom
        self.assertEqual(str(q.in_units_of(u.md_kilojoule/u.mole/u.nanometer)), '41.84 kJ/(nm mol)')

    def testQuantityComparisons(self):
        """ Tests binary comparison operators between Quantity """
        l1 = 1.0 * u.meters
        l2 = 2.0 * u.meters
        l3 = 1.0 * u.meters
        self.assertEqual(l1, l3)
        self.assertNotEqual(l1, l2)
        self.assertLessEqual(l1, l2)
        self.assertLess(l1, l2)
        self.assertGreater(l2, l1)
        self.assertGreaterEqual(l2, l1)
        self.assertLessEqual(l1, l3)
        self.assertGreaterEqual(l1, l3)
        self.assertGreater(1.2*u.meters, 72*u.centimeters)
        self.assertLess(1*u.meters, 200*u.centimeters)

    def testChemistryProblems(self):
        """ Tests some gen-chem applications with Quantity's """
        def work(f, dx):
            return f * dx

        F = 1.0 * u.kilogram * u.meter / u.second**2
        dx = 1.0 * u.meter
        self.assertEqual(work(F, dx), 1.0 * u.joule)
        self.assertEqual(F, 1.0 * u.newton)

        def ideal_gas_law(P, V, T):
            R = u.MOLAR_GAS_CONSTANT_R
            return (P * V / (R * T)).in_units_of(u.mole)

        T = (273.0 + 37.0) * u.kelvin
        P = (1.01325e5) * u.pascals
        r = 0.5e-6 * u.meters
        V = 4/3 * math.pi * r**3
        n = ideal_gas_law(P, V, T)
        val = 4/3*math.pi*0.5e-6**3*1
        self.assertAlmostEqualQuantities(P*V, val * u.atmospheres*u.meters**3)
        self.assertAlmostEqualQuantities(n, 2.05834818672e-17 * u.mole)
        self.assertAlmostEqualQuantities(V, 5.2359833333333e-19 * u.meters**3)
        self.assertEqual(str(T), '310.0 K')
        self.assertEqual(str(u.MOLAR_GAS_CONSTANT_R.format('%.4f')), '8.3145 J/(K mol)')
        self.assertTrue(u.is_quantity(V))
        # Checks trouble with complicated unit conversion factors
        p1 = 1.0 * u.atmospheres
        p2 = p1.in_units_of(u.joules/u.nanometers**3)
        V = 2.4 * u.nanometers**3
        beta = 4.e-4 * u.mole/u.joule
        x1 = beta * p1 * V
        y1 = x1 * u.AVOGADRO_CONSTANT_NA
        self.assertAlmostEqual(y1, 0.0585785776197)
        x2 = beta * p2 * V
        y2 = x2 * u.AVOGADRO_CONSTANT_NA
        self.assertAlmostEqual(y1, y2)

    def testAngleQuantities(self):
        """ Tests angle measurements """
        self.assertEqual(1.0*u.radians / u.degrees, 180 / math.pi)
        self.assertTrue(u.is_quantity(1.0*u.radians))
        self.assertTrue(u.is_quantity(1.0*u.degrees))
        self.assertEqual((1.0*u.radians).in_units_of(u.degrees), (180 / math.pi)*u.degrees)
        self.assertEqual(90*u.degrees/u.radians, math.pi/2)
        q = 90 * u.degrees + 0.3 * u.radians
        self.assertEqual(q._value, 90 + 180*0.3/math.pi)
        self.assertEqual(q.unit, u.degrees)

    def testIsQuantity(self):
        """ Tests if is_quantity can detect Quantities vs. scalars and units """
        self.assertTrue(u.is_quantity(1/u.second))
        self.assertFalse(u.is_quantity(u.second**-1))
        self.assertTrue(u.is_quantity(10*u.meters))
        self.assertFalse(u.is_quantity(10))

    def testBadQuantityMaths(self):
        """ Tests that Maths of incompatible units properly fails """
        self.assertRaises(TypeError, lambda:1*u.meters + 1*u.liters)
        self.assertRaises(AttributeError, lambda: 1*u.liters + 5)

    def testBadQuantityComparisons(self):
        """ Checks that comparisons of incompatible units fails """
        self.assertRaises(TypeError, lambda: 1*u.meters > 1*u.liters)
        self.assertRaises(TypeError, lambda: 1*u.nanometers > 1*u.degrees)

    def testDimensionless(self):
        """ Tests the properties of unit.dimensionless """
        x = 5 * u.dimensionless
        y = u.Quantity(5, u.dimensionless)
        self.assertTrue(u.is_quantity(x))
        self.assertTrue(u.is_quantity(y))
        self.assertNotEqual(x, 5)
        self.assertNotEqual(y, 5)
        self.assertEqual(x, y)
        self.assertEqual(x.value_in_unit_system(u.si_unit_system), 5)
        self.assertEqual(x.value_in_unit_system(u.cgs_unit_system), 5)
        self.assertEqual(x.value_in_unit_system(u.md_unit_system), 5)
        x = u.Quantity(1.0, u.dimensionless)
        y = u.Quantity(1.0, u.dimensionless)
        self.assertIsNot(x, y)
        self.assertEqual(x, y)

    def testTruthiness(self):
        """ Tests the truthiness of units """
        true = 2.3 * u.meters
        false = 0.0 * u.meters
        self.assertTrue(true)
        self.assertFalse(false)
        self.assertTrue(bool(true))
        self.assertFalse(bool(false))

    def testUnaryOperators(self):
        """ Tests unary operators on units """
        self.assertEqual(-(2.3*u.meters), u.Quantity(-2.3, u.meters))
        self.assertEqual(-(2.3*u.meters), -u.Quantity(2.3, u.meters))
        self.assertEqual(+(2.3*u.meters), u.Quantity(2.3, u.meters))
        self.assertEqual(2.3*u.meters, +u.Quantity(2.3, u.meters))

    def testUnitMathModule(self):
        """ Tests the unit_math functions on Quantity objects """
        self.assertEqual(u.sqrt(1.0*u.kilogram*u.joule), 1.0*u.kilogram*u.meter/u.second)
        self.assertEqual(u.sqrt(1.0*u.kilogram*u.calorie), math.sqrt(4.184)*u.kilogram*u.meter/u.second)
        self.assertEqual(u.sqrt(9), 3) # Test on a scalar
        self.assertEqual(u.sin(90*u.degrees), 1)
        self.assertEqual(u.sin(math.pi/2*u.radians), 1)
        self.assertEqual(u.sin(math.pi/2), 1)
        self.assertEqual(u.cos(180*u.degrees), -1)
        self.assertEqual(u.cos(math.pi*u.radians), -1)
        self.assertEqual(u.cos(math.pi), -1)
        self.assertAlmostEqual(u.tan(45*u.degrees), 1)
        self.assertAlmostEqual(u.tan(math.pi/4*u.radians), 1)
        self.assertAlmostEqual(u.tan(math.pi/4), 1)
        acos = u.acos(1.0)
        asin = u.asin(1.0)
        atan = u.atan(1.0)
        self.assertTrue(u.is_quantity(acos))
        self.assertTrue(u.is_quantity(asin))
        self.assertTrue(u.is_quantity(atan))
        self.assertEqual(acos.unit, u.radians)
        self.assertEqual(asin.unit, u.radians)
        self.assertEqual(atan.unit, u.radians)
        self.assertEqual(acos.value_in_unit(u.degrees), 0)
        self.assertEqual(acos / u.radians, 0)
        self.assertEqual(asin.value_in_unit(u.degrees), 90)
        self.assertEqual(asin / u.radians, math.pi/2)
        self.assertAlmostEqual(atan.value_in_unit(u.degrees), 45)
        self.assertAlmostEqual(atan / u.radians, math.pi/4)
        # Check some sequence maths
        seq = [1, 2, 3, 4] * u.meters
        self.assertEqual(u.sum(seq), 10*u.meters)
        self.assertEqual(u.dot(seq, seq), (1+4+9+16)*u.meters**2)
        self.assertEqual(u.norm(seq), math.sqrt(30)*u.meters)

    def testUnitMathModuleBadInput(self):
        """ Tests that bad units to unit_math fails appropriately """
        self.assertRaises(ArithmeticError, lambda: u.sqrt(9*u.meters))
        self.assertRaises(TypeError, lambda: u.sin(1*u.meters))

    def testBadQuantityConversions(self):
        """ Tests that conversions to incompatible units fails """
        with self.assertRaises(TypeError):
            (1.0*u.meters).in_units_of(u.seconds)
        with self.assertRaises(TypeError):
            (1.0*u.meters).value_in_unit(u.seconds)

    def testCreateNewBaseUnit(self):
        """ Tests creating a new base unit """
        ms = u.milli * u.second_base_unit
        self.assertIsInstance(ms, u.BaseUnit)
        self.assertEqual(repr(ms), 'BaseUnit(base_dim=BaseDimension("time"), name="millisecond", symbol="ms")')
        self.assertEqual(ms.conversion_factor_to(u.second_base_unit), 0.001)

    def testCreateNewScaledUnit(self):
        """ Tests creating a new ScaledUnit """
        mC = u.milli * u.ScaledUnit(4.184, u.joule, "calorie", "cal")
        self.assertIsInstance(mC, u.ScaledUnit)
        self.assertEqual(repr(mC), "ScaledUnit(factor=0.004184, master=joule, name='millicalorie', symbol='mcal')")
        self.assertFalse(u.is_unit(mC))

    def testCreateNewUnit(self):
        """ Tests creating a new Unit """
        ms = u.milli * u.second
        self.assertTrue(u.is_unit(ms))
        self.assertEqual(repr(ms),
                'Unit({BaseUnit(base_dim=BaseDimension("time"), '
                'name="millisecond", symbol="ms"): 1.0})')

    def testIllegalQuantityOps(self):
        """ Test that illegal operations on Quantity objects fails """
        self.assertRaises(TypeError, lambda: u.milli * (1.0 * u.second))

    def testQuantityFormat(self):
        """ Tests the format method on Quantity instances """
        x = 5.439999999 * u.picosecond
        self.assertEqual(x.format("%.3f"), '5.440 ps')

    def testQuantityCollectionMethods(self):
        """ Tests some sequence methods on Quantity objects (no numpy) """
        seq = [1, 2, 3, 4] * u.meters
        self.assertEqual(seq.sum(), 10*u.meters)
        self.assertEqual(seq.mean(), 2.5*u.meters)
        self.assertEqual(seq.max(), 4*u.meters)
        self.assertEqual(seq.min(), 1*u.meters)
        self.assertAlmostEqualQuantities(seq.std(), 1.1180339887498949*u.meters)

    def testString(self):
        """ Tests unit handling with strings, which should be dimensionless """
        s = u.Quantity("string")
        self.assertEqual(s.value_in_unit_system(u.md_unit_system), "string")

    def testMisc(self):
        """ Miscellaneous tests for the unit package """
        self.assertTrue(u.meter is not None)
        self.assertFalse(u.meter is None)
        self.assertEqual(repr(1.2*u.meters), 'Quantity(value=1.2, unit=meter)')
        class Foo(object):
            def bar(self):
                return 'bar'
        x = Foo()
        self.assertEqual(x.bar(), 'bar')
        y = x * u.nanometers
        self.assertEqual(y.bar(), 'bar')
        self.assertEqual(str(u.meters*u.centimeters), 'centimeter*meter')
        self.assertEqual(str(u.meters*u.meters), 'meter**2')
        self.assertEqual(str(u.meter*u.meter), 'meter**2')

class TestNumpyUnits(utils.QuantityTestCase):

    def testNumpyQuantity(self):
        """ Tests that numpy arrays can form Quantity values """
        q = u.Quantity(np.array([1, 2, 3]), u.centimeters)
        self.assertTrue(u.is_quantity(q))
        self.assertIsInstance(q._value, np.ndarray)
        self.assertTrue(np.all(q / u.millimeters == np.array([1, 2, 3])*10))

    def testNumpyDeepCopy(self):
        """ Check that deepcopy on numpy array does not strip units """
        x = u.Quantity(np.zeros((2, 3)), u.nanometer)
        y = copy.deepcopy(x)
        self.assertTrue(np.all(x == y))
        self.assertTrue(u.is_quantity(x))
        self.assertTrue(u.is_quantity(y))

    def testNumpyDivision(self):
        """ Tests that division of numpy Quantities works correctly """
        x = u.Quantity(np.asarray([1., 2.]), u.nanometers)
        y = u.Quantity(np.asarray([3., 4.]), u.picoseconds)
        xy = x / y
        self.assertTrue(u.is_quantity(xy))
        self.assertEqual(xy.unit, u.nanometers/u.picoseconds)
        self.assertEqual(xy[0].value_in_unit(u.nanometers/u.picoseconds), 1/3)
        self.assertEqual(xy[1].value_in_unit(u.nanometers/u.picoseconds), 0.5)

    def testNumpyIsString(self):
        """ Tests the internal _is_string method with numpy Quantities """
        a = np.array([[1, 2, 3], [4, 5, 6]])
        self.assertIsInstance("", str)
        self.assertTrue(isinstance("", str))
        self.assertTrue(isinstance("t", str))
        self.assertTrue(isinstance("test", str))
        self.assertFalse(isinstance(3, str))
        self.assertFalse(isinstance(a, str))

    def testNumpyFunctions(self):
        """ Tests various numpy attributes that they result in Quantities """
        a = u.Quantity(np.arange(10), u.seconds)
        self.assertEqual(a.max(), 9*u.seconds)
        self.assertEqual(a.min(), 0*u.seconds)
        self.assertEqual(a.mean(), 4.5*u.seconds)
        self.assertAlmostEqualQuantities(a.std(), 2.8722813232690143*u.seconds)
        b = a.reshape((5, 2))
        self.assertTrue(u.is_quantity(b))

    def testMultiplication(self):
        """ Tests that units override numpy.ndarray multiplication """
        self.assertIsInstance(np.arange(10)*u.angstroms, u.Quantity)
        x = np.array([1]) * u.liters
        self.assertIsInstance(x, u.Quantity)
        self.assertIsInstance(np.arange(10) * x, u.Quantity)
